"""
Labour market overview, UK: December 2020 - Real Survey Data Example
--------------------------------------------------------------------
This example demonstrates how to replicate the Labour Market overview accessible
example found at https://analysisfunction.civilservice.gov.uk/policy-store/further-resources-for-releasing-statistics-in-spreadsheets/
"""

from pathlib import Path

import pandas as pd

import gptables as gpt

# Read data and arrange
parent_dir = Path(__file__).parent

labour_market_data = pd.read_csv(parent_dir / "survey_data.csv")
labour_market_data.dropna(
    axis=0, how="all", inplace=True
)  # Remove empty rows in the data
labour_market_data.dropna(
    axis=1, how="all", inplace=True
)  # Remove columns rows in the data
col_names = [
    "Time period and dataset code row",
    "Number of people",
    "Economically active",
    "Employment level",
    "Unemployment level",
    "Economically inactive",
    "Economically active rate",
    "Employment rate",
    "Unemployment rate",
    "Economically inactive rate",
]
labour_market_data.columns = col_names


# Define table elements
table_name = "Labour_market_overview_accessibility_example_Nov21"
title = "Number and percentage of population aged 16 and over in each labour market activity group, UK, seasonally adjusted"
subtitles = [
    "This worksheet contains one table. Some cells refer to notes which can be found on the notes worksheet."
]
units = {
    1: "thousands",
    2: "thousands",
    3: "thousands",
    4: "thousands",
    5: "thousands",
    6: "%",
    7: "%",
    8: "%",
    9: "%",
}
table_notes = {
    2: "$$note 1$$",
    3: "$$note 2$$",
    4: "$$note 2$$",
    5: "$$note 3$$",
    7: "$$note 4$$",
    8: "$$note 4$$",
    9: "$$note 4$$",
}
scope = "Labour Market"
source = "Source: Office for National Statistics"
index = {2: 0}  # Column 0 is a level 2 index
additional_formatting = [
    {
        "row": {
            "rows": [1],
            "format": {"bold": True, "font_size": 14},
        }
    }
]


# or use kwargs to pass these to the appropriate parameters
kwargs = {
    "table_name": table_name,
    "title": title,
    "subtitles": subtitles,
    "units": units,
    "table_notes": table_notes,
    "scope": scope,
    "source": source,
    "index_columns": index,
    "additional_formatting": additional_formatting,
}


# Define our GPTable
survey_table = gpt.GPTable(table=labour_market_data, **kwargs)

sheets = {"sheet 1a": survey_table}

cover = gpt.Cover(
    cover_label="Cover",
    title="Labour market overview data tables, UK, December 2020 (accessibility example)",
    intro=[
        "This spreadsheet contains a selection of the data tables published alongside the Office for National Statistics' Labour market overview for December 2020. We have edited these data tables and the accompanying cover sheet, table of contents and notes worksheet to meet the legal accessibility regulations. It is intended to be an example of an accessible spreadsheet. The data tables and accompanying information have not been quality assured. Please see the original statistical release if you are looking for accurate data.",
        "[Labour market overview, UK: December 2020](https://www.ons.gov.uk/employmentandlabourmarket/peopleinwork/employmentandemployeetypes/bulletins/uklabourmarket/december2020)",
    ],
    about=[
        [{"bold": True, "font_size": 14}, "Publication dates"],
        "This data tables in this spreadsheet were originally published at 7:00am 15 December 2020",
        "The next publication was published at 7:00am 26 January 2021.",
        [{"bold": True, "font_size": 14}, "Note on weighting methodology"],
        "Due to the coronavirus (COVID19) pandemic, all face to face interviewing for the Labour Force Survey was suspended and replaced with telephone interviewing. This change in mode for first interviews has changed the non-response bias of the survey, affecting interviews from March 2020 onwards. All data included in this spreadsheet have now been updated and are based on latest weighting methodology.",
        "More information about the impact of COVID19 on the Labour Force Survey",
        "Dataset identifier codes",
        "The four-character identification codes appearing in the tables are the ONS' references for the data series.",
        [{"bold": True, "font_size": 14}, "Comparing quarterly changes"],
        "When comparing quarterly changes ONS recommends comparing with the previous non-overlapping three-month average time period, for example, compare Apr to Jun with Jan to Mar, not with Mar to May.",
        [{"bold": True, "font_size": 14}, "Units, notes and no data"],
        "Some cells in the tables refer to notes which can be found in the notes worksheet. Note markers are presented in square brackets, for example: [note 1].",
        "Some cells have no data, when this is the case the words 'no data' are presented in square brackets, for example: '[no data]'. An explanation of why there is no data is given in the notes worksheet, see the column headings for which notes you should refer to.",
        "Some column headings give units, when this is the case the units are presented in round brackets to differentiate them from note markers.",
        [
            {"bold": True, "font_size": 14},
            "Historic publication dates for labour market statistics",
            " ",
        ],
        "The monthly labour market statistics release was first published in April 1998. Prior to April 1998 there was no integrated monthly release and the Labour Force Survey estimates were published separately, on different dates, from other labour market statistics. From April 2018 the usual publication day for the release was changed from Wednesday to Tuesday.",
        [{"bold": True, "font_size": 14}, "More labour market data"],
        "Other labour market datasets are available on the ONS website.",
        "Labour market statistics time series dataset on the ONS website.",
    ],
    contact=[
        "Tel: 01633455400",
        "Email: [labour.market@ons.gov.uk](mailto:labour.market@ons.gov.uk)",
    ],
)

# Notesheet
notes_table = pd.read_csv(parent_dir / "survey_data_notes.csv")
notes_table.dropna(axis=0, how="all", inplace=True)  # Remove empty rows in the data
notes_table.dropna(axis=1, how="all", inplace=True)  # Remove columns rows in the data
notes_table.columns = ["Note reference", "Note text"]

# Use write_workbook to win!
if __name__ == "__main__":
    output_path = parent_dir / "python_survey_data_gptable.xlsx"
    gpt.write_workbook(
        filename=output_path,
        sheets=sheets,
        cover=cover,
        notes_table=notes_table,
        contentsheet_options={"additional_elements": ["subtitles", "scope"]},
        auto_width=True,
        gridlines="show_all",
        cover_gridlines=True,
    )
    print("Output written at: ", output_path)



import warnings
from pathlib import Path
from typing import Any, Dict, Optional, Union

import pandas as pd

from gptables.core.cover import Cover
from gptables.core.gptable import GPTable
from gptables.core.theme import Theme
from gptables.core.wrappers import GPWorkbook


def produce_workbook(
    filename: str,
    sheets: Dict[str, "GPTable"],
    theme: Optional["Theme"] = None,
    cover: Optional["Cover"] = None,
    contentsheet_label: str = "Contents",
    contentsheet_options: Optional[Dict[str, Any]] = None,
    notes_table: Optional[pd.DataFrame] = None,
    notesheet_label: str = "Notes",
    notesheet_options: Optional[Dict[str, Any]] = None,
    auto_width: Union[bool, Dict[str, bool]] = True,
    gridlines: str = "hide_all",
    cover_gridlines: bool = False,
) -> GPWorkbook:
    """
    Produces a formatted workbook.

    Can be written to an `.xlsx` file as specified in `filename` using `.close()`.

    Parameters
    ----------
    filename : str
        path to write final workbook to (an `.xlsx` file)
    sheets : dict
        mapping worksheet labels to ``GPTable`` objects
    theme : gptables.Theme, optional
        formatting to be applied to GPTable elements. ``gptheme`` is used by
        default.
    cover : gptables.Cover, optional
        cover page text. Including this argument will generate a cover page.
    contentsheet_label : str
        table of contents sheet label, defaults to "Contents". If None, table
        of contents will not be generated.
    contentsheet_options : dict, optional
        dictionary of contentsheet customisation parameters. Valid keys are
        `additional_elements`, `column_names`, `table_name`, `title`,
        `subtitles` and `instructions`.
    notes_table : pd.DataFrame, optional
        table with notes reference, text and (optional) link columns. If None,
        notes sheet will not be generated.
    notesheet_label : str, optional
        notes sheet label, defaults to "Notes"
    notesheet_options : dict, optional
        dictionary of notesheet customisation parameters. Valid keys are
        `table_name`, `title` and `instructions`.
    auto_width : bool or dict, optional
        If bool, applies to all sheets. If dict, should map sheet labels to bools.
    gridlines : string, optional
        option to hide or show gridlines on worksheets. "show_all" - don't
        hide gridlines, "hide_printed" - hide printed gridlines only, or
        "hide_all" - hide screen and printed gridlines.
    cover_gridlines : bool, optional
        indication if gridlines should apply to the cover worksheet. False
        by default.

    Returns
    -------
    workbook : gptables.GPWorkbook
    """
    if contentsheet_options is None:
        contentsheet_options = {}
    if notesheet_options is None:
        notesheet_options = {}

    if isinstance(filename, Path):
        filename = filename.as_posix()

    wb = GPWorkbook(filename)

    if theme is not None:
        wb.set_theme(theme)

    if cover is not None:
        if cover_gridlines:
            ws = wb.add_worksheet(cover.cover_label, gridlines=gridlines)
        else:
            ws = wb.add_worksheet(cover.cover_label, gridlines="hide_all")
        ws.write_cover(cover)

    contentsheet = {}
    if contentsheet_label is not None:
        if contentsheet_options:
            valid_keys = [
                "additional_elements",
                "column_names",
                "table_name",
                "title",
                "subtitles",
                "instructions",
            ]
            if not all(key in valid_keys for key in contentsheet_options.keys()):
                msg = (
                    "Valid `contentsheet_options` keys are 'additional_elements',"
                    "'column_names', 'table_name', 'title', 'subtitles', 'instructions'"
                )
                raise ValueError(msg)
        contents_gptable = wb.make_table_of_contents(sheets, **contentsheet_options)
        contentsheet = {contentsheet_label: contents_gptable}

    wb._update_annotations(sheets)

    notesheet = {}
    if notes_table is None:
        warnings.warn("No note text provided, notes sheet has not been generated")
    else:
        note_gptable = wb.make_notesheet(notes_table, **notesheet_options)
        notesheet = {notesheet_label: note_gptable}

    sheets = {**contentsheet, **notesheet, **sheets}
    for label, gptable in sheets.items():
        ws = wb.add_worksheet(label, gridlines=gridlines)
        if isinstance(auto_width, dict):
            sheet_auto_width = auto_width.get(label, True)
        else:
            sheet_auto_width = auto_width
        ws.write_gptable(gptable, sheet_auto_width, wb._annotations)

    return wb


def write_workbook(
    filename: str,
    sheets: Dict[str, "GPTable"],
    theme: Optional["Theme"] = None,
    cover: Optional["Cover"] = None,
    contentsheet: Optional[str] = None,
    contentsheet_label: str = "Contents",
    contentsheet_options: Optional[Dict[str, Any]] = None,
    notes_table: Optional[pd.DataFrame] = None,
    notesheet_label: str = "Notes",
    notesheet_options: Optional[Dict[str, Any]] = None,
    auto_width: Union[bool, Dict[str, bool]] = True,
    gridlines: str = "hide_all",
    cover_gridlines: bool = False,
) -> None:
    """
    Writes a formatted Excel workbook to `filename`.

    Parameters
    ----------
    filename : str
        Path to write final workbook to (an `.xlsx` file)
    sheets : dict
        mapping worksheet labels to ``GPTable`` objects
    theme : gptables.Theme, optional
        formatting to be applied to GPTable elements. ``gptheme`` is used by
        default.
    cover : gptables.Cover, optional
        cover page text. Including this argument will generate a cover page.
    contentsheet_label : str
        table of contents sheet label, defaults to "Contents". If None, table
        of contents will not be generated.
    contentsheet_options : dict, optional
        dictionary of contentsheet customisation parameters. Valid keys are
        `additional_elements`, `column_names`, `table_name`, `title`,
        `subtitles` and `instructions`
    notes_table : pd.DataFrame, optional
        table with notes reference, text and (optional) link columns. If None,
        notes sheet will not be generated.
    notesheet_label : str, optional
        notes sheet label, defaults to "Notes"
    notesheet_options : dict, optional
        dictionary of notesheet customisation parameters. Valid keys are
        `table_name`, `title` and `instructions`.
    auto_width : bool, optional
        indicate if column widths should be automatically determined. True by
        default.
    gridlines : string, optional
        option to hide or show gridlines on worksheets. "show_all" - don't
        hide gridlines, "hide_printed" - hide printed gridlines only, or
        "hide_all" - hide screen and printed gridlines.
    cover_gridlines : bool, optional
        indication if gridlines should apply to the cover worksheet. False
        by default.

    Returns
    -------
    None
    """

    if contentsheet_options is None:
        contentsheet_options = {}
    if notesheet_options is None:
        notesheet_options = {}

    wb = produce_workbook(
        filename,
        sheets,
        theme,
        cover,
        contentsheet_label,
        contentsheet_options,
        notes_table,
        notesheet_label,
        notesheet_options,
        auto_width,
        gridlines,
        cover_gridlines,
    )
    wb.close()



from pathlib import Path

import pandas as pd

import gptables as gpt

parent_dir = Path(__file__).parents[1]
penguins_data = pd.read_csv(parent_dir / "test/data/penguins.csv")

formatted_subtitles = [
    "The first subtitle",
    [{"bold": True}, "This", " is another subtitle"],
]

sample_additional_formatting = [
    {
        "column": {
            "columns": ["Species", "Island"],
            "format": {
                "align": "center",
                "italic": True,
            },
        }
    },
    {"column": {"columns": [3], "format": {"left": 1}}},
    {
        "row": {
            "rows": -1,
            "format": {
                "bottom": 1,
                "indent": 2,
            },
        }
    },
]

penguins_table = gpt.GPTable(
    table=penguins_data,
    table_name="penguins_statistics",
    title="The Palmer Penguins Dataset",
    subtitles=formatted_subtitles,
    scope="Penguins",
    source="Palmer Station, Antarctica",
    additional_formatting=sample_additional_formatting,
)

penguins_sheets = {"Penguins": penguins_table}

if __name__ == "__main__":
    output_path = parent_dir / "gpt_additional_formatting.xlsx"
    wb = gpt.produce_workbook(filename=output_path, sheets=penguins_sheets)
    wb.close()
    print("Output written at: ", output_path)



from contextlib import redirect_stdout
from itertools import chain, combinations

import importlib_resources
import pytest

from gptables import Theme, gptheme

valid_description_elements = ["legend", "instructions", "scope", "source"]


def powerset(iterable):
    s = list(iterable)
    return chain.from_iterable(combinations(s, r) for r in range(len(s) + 1))


@pytest.fixture()
def empty_theme():
    yield Theme()


class TestCleanInitTheme:
    """
    Test initialisation of the Theme class without config.
    """

    @pytest.mark.parametrize("attr", Theme()._format_attributes)
    def test_default_format_attrs(self, attr, empty_theme):
        """Test Theme attribute default types"""
        exp = {}
        got = getattr(empty_theme, attr)
        assert exp == got

    def test_default_other_attrs(self, empty_theme):
        assert empty_theme.description_order == []

    def test_print_attributes(self, empty_theme):
        """Test Theme print_attributes()"""
        from io import StringIO

        file_handler = StringIO()
        with redirect_stdout(file_handler):
            empty_theme.print_attributes()

        got = file_handler.getvalue()

        exp = """cover_title_format : {}
cover_subtitle_format : {}
cover_text_format : {}
title_format : {}
subtitle_format : {}
instructions_format : {}
scope_format : {}
column_heading_format : {}
index_1_format : {}
index_2_format : {}
index_3_format : {}
data_format : {}
source_format : {}
legend_format : {}
description_order : []
"""

        assert got == exp


class TestConfigInitTheme:
    """
    Test initialisation of the Theme class using a config dictionary.
    """

    def test_dict_init(self):
        config = {
            "global": {
                "font_size": 12,
                "font_name": "Arial",
                "font_color": "automatic",
            },
            "cover_title": {
                "font_size": 16,
                "bold": True,
                "text_wrap": True,
            },
            "cover_subtitle": {
                "font_size": 14,
                "bold": True,
                "text_wrap": True,
            },
            "cover_text": {
                "text_wrap": True,
            },
            "title": {"bold": True, "font_size": 16},
            "subtitle": {"font_size": 14},
            "instructions": None,
            "scope": None,
            "column_heading": {
                "bold": True,
                "bottom": 1,
                "text_wrap": 1,
                "valign": "top",
            },
            "index_1": {"bold": True, "text_wrap": 1},
            "index_2": {"text_wrap": 1},
            "index_3": {"text_wrap": 1},
            "data": {"text_wrap": 1},
            "source": {"font_size": 12},
            "legend": {"font_size": 12},
            "description_order": ["instructions", "legend", "source", "scope"],
        }
        got = Theme(config)

        exp = gptheme

        assert exp == got

    def test_file_init(self):
        """
        Test initialisation of Theme using default theme yaml config file.
        """
        config_file = importlib_resources.files("gptables") / "themes/gptheme.yaml"

        got = Theme(str(config_file))

        exp = gptheme

        assert exp == got


class TestFormatValidationTheme:
    """
    Test validation of format dictionaries.
    """

    def test_invalid_attribute_config(self, empty_theme):
        """
        Test that invalid attribute names in config raises a ValueError.
        """
        config = {"potato": {"font_size": 9}}
        with pytest.raises(ValueError):
            empty_theme.apply_config(config)

    def test_invalid_format_label_config(self, empty_theme):
        """
        Test that invalid format labels in config raises a ValueError.
        """
        config = {"notes": {"not_a_format": 5}}
        with pytest.raises(ValueError):
            empty_theme.apply_config(config)

    @pytest.mark.parametrize("attr", Theme()._format_attributes)
    def test_invalid_format_label_single_attr(self, attr, empty_theme):
        """
        Test that `validate_single_format` decorator catches invalid format
        labels in individual format dictionaries.
        """
        format_dict = {"not_a_format": 5}
        with pytest.raises(ValueError):
            getattr(empty_theme, "update_" + attr)(format_dict)

    @pytest.mark.parametrize("attr", Theme()._format_attributes)
    def test_valid_format_label(self, attr, empty_theme):
        """
        Test that Theme attribute is correctly updated when valid format label
        is used.
        """
        getattr(empty_theme, "update_" + attr)({"font_size": 9})

        exp = {"font_size": 9}
        got = getattr(empty_theme, attr)

        assert exp == got

    def test_valid_format_label_config(self, empty_theme):
        """
        Test that valid format labels in config changes specified format attr
        but not an unrelated attr. Previous bug updated all formats whenever
        one was updated.
        """
        config = {"source": {"font_size": 5}}
        empty_theme.apply_config(config)

        exp = {"font_size": 5}
        got = empty_theme.source_format

        assert exp == got

        got2 = empty_theme.title_format

        assert {} == got2


class TestOtherValidationTheme:
    """
    Test validation of non-format Theme attributes.
    """

    @pytest.mark.parametrize(
        "format_order", ["notes", {"annotations": 2}, 1, 3.14, False, None]
    )
    def test_invalid_description_order_type(self, format_order, empty_theme):
        """
        Test that non-list description_order entries raise a TypeError.
        """
        with pytest.raises(TypeError):
            empty_theme.update_description_order(format_order)

    @pytest.mark.parametrize("format_order", [["potato"], [1], [3.14], [dict()], [[]]])
    def test_invalid_description_order_values(self, format_order, empty_theme):
        """
        Test that list description_order entries containing invalid elements raises
        a ValueError.
        """
        with pytest.raises(ValueError):
            empty_theme.update_description_order(format_order)

    @pytest.mark.parametrize("description_order", powerset(valid_description_elements))
    def test_valid_description_order_values(self, description_order, empty_theme):
        """
        Test that valid list description_order entries are used to set
        the corresponding attribute.
        """
        empty_theme.update_description_order(list(description_order))
        assert getattr(empty_theme, "description_order") == list(description_order)



from pathlib import Path

import pandas as pd
import pytest

import gptables as gpt
from gptables.test.test_utils.excel_comparison_test import ExcelComparisonTest


@pytest.fixture(scope="function")
def create_gpworkbook():

    def generate_gpworkbook(output_path):
        """
        Create GPWorkbook to be used with API tests.
        """
        table = pd.DataFrame({"columnA": ["x", "y"], "columnB": [0, 1]})

        gptable = gpt.GPTable(
            table=table,
            table_name="table_name",
            title="Title$$ref1$$",
            subtitles=["Subtitle1", "Subtitle2"],
            instructions=["Some ", {"bold": True}, "instructions"],
            legend=["s = symbol", "t = shorthand"],
            scope="Tests",
            source="My imagination",
            units={0: "Latin alphabet", "columnB": "real numbers"},
            table_notes={1: "$$ref2$$"},
            additional_formatting=[
                {"column": {"columns": ["columnA"], "format": {"bold": True}}}
            ],
        )

        notes_table = pd.DataFrame(
            {
                "Note reference": ["ref1", "ref2"],
                "Note text": ["Some text", "Some more text"],
                "Link": [
                    "[gov.uk](https://www.gov.uk)",
                    "[Wikipedia](https://en.wikipedia.org)",
                ],
            }
        )

        cover = gpt.Cover(
            title="Cover title",
            intro=["Introduction"],
            about=["About"],
            contact=["Me", "[please.dont@contact.me](mailto:please.dont@contact.me)"],
            cover_label="Cover",
        )

        gpt.write_workbook(  # Use defaults for theme and autowidth
            filename=output_path / "test_end_to_end.obtained.xlsx",
            sheets={"Label": gptable},
            cover=cover,
            contentsheet_label="Table of contents",
            contentsheet_options={"additional_elements": ["subtitles"]},
            notes_table=notes_table,
            notesheet_label="Notes table",
            notesheet_options={"title": "Table with notes"},
            gridlines="show_all",
            cover_gridlines=False,
        )

    return generate_gpworkbook


def test_end_to_end(create_gpworkbook, file_regression):
    """
    Test that runs the API functions with example input and checks the generated Excel
    file using a manually managed expected_workbook.xlsx and ExcelComparisonTest.
    """
    output_path = Path(__file__).parent / "test_api"
    output_path.mkdir(parents=True, exist_ok=True)
    create_gpworkbook(output_path)
    actual_file = output_path / "test_end_to_end.obtained.xlsx"

    # Use file_regression to manage the regression file, but ignore its assertion
    try:
        file_regression.check(actual_file.read_bytes(), extension=".xlsx", binary=True)
    except AssertionError:
        pass

    regression_file = output_path / "test_end_to_end.xlsx"
    if not regression_file.exists():
        raise FileNotFoundError(
            f"Could not find regression file for comparison: {regression_file}"
        )

    ect = ExcelComparisonTest()
    ect.got_filename = str(actual_file)
    ect.exp_filename = str(regression_file)
    ect.ignore_files = []
    ect.ignore_elements = {}
    ect.assertExcelEqual()
    ect.tearDown()



from collections import namedtuple

import pandas as pd
import pytest
import xlsxwriter
from pandas.testing import assert_frame_equal, assert_series_equal

import gptables
from gptables import Theme, gptheme
from gptables.core.gptable import FormatList
from gptables.core.wrappers import GPWorkbook, GPWorksheet
from gptables.test.test_gptable import does_not_raise

Tb = namedtuple("Testbook", "wb ws")

valid_text_elements = [  # Not None
    "This is a string",
    FormatList(["More than ", {"italic": True}, "just ", "a string"]),
    FormatList([{"bold": True}, "text"]),
]

test_text_list = [
    "This has a $$reference$$",
    "This one doesn't",
    "Here's $$another$$one",
]

exp_text_list = ["This has a [note 1]", "This one doesn't", "Here's one[note 2]"]


invalid_text_elements = [dict(), set(), 42, 3.14, True]


@pytest.fixture()
def testbook():
    # See https://github.com/jmcnamara/XlsxWriter/issues/746#issuecomment-685869888
    wb = GPWorkbook(options={"in_memory": True})
    ws = wb.add_worksheet()
    yield Tb(wb, ws)
    wb.fileclosed = 1


class TestGPWorksheetInit:
    """
    Test that default attributes are set when GPWorksheets are created.
    """

    def test_subclass(self):
        """
        Test that the GPWorksheet class is a subclass of the XlsxWriter
        Worksheet class.
        """
        assert issubclass(GPWorksheet, xlsxwriter.worksheet.Worksheet)

    def test_default_theme_set(self, testbook):
        """
        Test that the default theme (gptheme) is used when no theme is set.
        """
        assert testbook.wb.theme == gptheme

    def test_default_gridlines(self, testbook):
        """
        Test that print and screen gridlines are hidden by default.
        """
        assert testbook.ws.print_gridlines == 0
        assert testbook.ws.screen_gridlines == 0

    def test_wb_reference(self, testbook):
        """
        Test that GPWorksheets reference their parent GPWorkbook.
        """
        assert testbook.ws._workbook == testbook.wb

    @pytest.mark.parametrize(
        "not_a_gptable", [dict(), set(), [], 1, 3.14, "test_string", pd.DataFrame()]
    )
    def test_invalid_write_gptable(self, not_a_gptable, testbook):
        """
        Test that write_gptable() raises a TypeError when argument is not a
        gptables.GPTable object.
        """
        with pytest.raises(TypeError):
            testbook.ws.write_gptable(not_a_gptable)


class TestGPWorksheetWriting:
    """
    Test that additional writing methods correctly write to GPWorksheet object.
    """

    def test__smart_write_str(self, testbook):
        """
        Test that strings are stored in the GPWorksheet as expected.
        """
        testbook.ws._smart_write(0, 0, valid_text_elements[0], {})
        # Strings are stored in a lookup table for efficiency
        got_string = testbook.ws.str_table.string_table
        exp_string = {valid_text_elements[0]: 0}
        assert got_string == exp_string

        # String is referenced using a named tuple (string, Format)
        # Here we get first element, which references string lookup location
        got_lookup = testbook.ws.table[0][0][0]
        exp_lookup = 0
        assert got_lookup == exp_lookup

    def test__smart_write_formatted_str(self, testbook):
        testbook.ws._smart_write(1, 2, valid_text_elements[0], {"bold": True})
        # Strings are stored in a lookup table for efficiency
        got_string = testbook.ws.str_table.string_table
        exp_string = {valid_text_elements[0]: 0}
        assert got_string == exp_string

        # String is referenced using a named tuple (string, Format)
        # Here we get first element, which references string lookup location
        cell = testbook.ws.table[1][2]
        got_lookup = cell[0]
        exp_lookup = 0
        assert got_lookup == exp_lookup

        format_obj = cell[1]
        assert format_obj.bold

    def test__smart_write_rich_text(self, testbook):
        testbook.wb.set_theme(Theme({}))

        testbook.ws._smart_write(0, 0, valid_text_elements[1], {})
        # Strings are stored in a lookup table for efficiency
        got_string = testbook.ws.str_table.string_table
        exp_string = {
            '<r><t xml:space="preserve">More than </t></r><r><rPr><i'
            '/><sz val="11"/><color theme="1"/><rFont val="Calibri"/'
            '><family val="2"/><scheme val="minor"/></rPr><t xml:spa'
            'ce="preserve">just </t></r><r><rPr><sz val="11"/><color'
            ' theme="1"/><rFont val="Calibri"/><family val="2"/><sch'
            'eme val="minor"/></rPr><t>a string</t></r>': 0
        }
        assert got_string == exp_string

        # String is referenced using a named tuple (string, Format)
        # Here we get first element, which references string lookup location
        got_lookup = testbook.ws.table[0][0][0]
        exp_lookup = 0
        assert got_lookup == exp_lookup

    def test__smart_write_formatted_rich_text(self, testbook):
        testbook.wb.set_theme(Theme({}))

        testbook.ws._smart_write(1, 2, valid_text_elements[1], {})
        # Strings are stored in a lookup table for efficiency
        got_string = testbook.ws.str_table.string_table
        exp_string = {
            '<r><t xml:space="preserve">More than </t></r><r><rPr>'
            '<i/><sz val="11"/><color theme="1"/><rFont val="Calib'
            'ri"/><family val="2"/><scheme val="minor"/></rPr><t xml'
            ':space="preserve">just </t></r><r><rPr><sz val="11"'
            '/><color theme="1"/><rFont val="Calibri"/><family val="'
            '2"/><scheme val="minor"/></rPr><t>a string</t></r>': 0
        }

        assert got_string == exp_string

        # String is referenced using a named tuple (string, Format)
        # Here we get first element, which references string lookup location
        cell = testbook.ws.table[1][2]
        got_lookup = cell[0]
        exp_lookup = 0
        assert got_lookup == exp_lookup

    def test__smart_write_link(self, testbook):
        testbook.wb.set_theme(Theme({}))

        display_text = "gov.uk"
        url = "https://www.gov.uk/"

        testbook.ws._smart_write(0, 0, {display_text: url}, {})

        got_string = testbook.ws.str_table.string_table
        exp_string = {display_text: 0}
        assert got_string == exp_string

        got_hyperlink = testbook.ws.hyperlinks[0][0]._link
        exp_hyperlink = url
        assert got_hyperlink == exp_hyperlink

        # String is referenced using a named tuple (string, Format)
        # Here we get first element, which references string lookup location
        cell = testbook.ws.table[0][0]

        got_lookup = cell[0]
        exp_lookup = 0
        assert got_lookup == exp_lookup

        format_obj = cell[1]
        assert format_obj.underline is True
        assert format_obj.font_color._rgb_hex_value() == "0000FF"  # aka Blue

    def test__smart_write_null_cell(self, testbook):
        testbook.ws._smart_write(0, 0, None, {})
        # Strings are stored in a lookup table for efficiency
        got_string = testbook.ws.str_table.string_table
        exp_string = {}
        assert got_string == exp_string

        # Strings referenced using a named tuple (string, Format)
        # When cell has no content, tuple only contains Format
        cell = testbook.ws.table[0][0]
        assert len(cell) == 1

    def test__write_empty_table(self, testbook, create_gptable_with_kwargs):
        gptable = create_gptable_with_kwargs({"table": pd.DataFrame({"col": [None]})})
        with pytest.raises(ValueError):
            testbook.ws._write_table_elements([0, 0], gptable, auto_width=True)

    def test__write_integer_table(self, testbook, create_gptable_with_kwargs):
        gptable = create_gptable_with_kwargs(
            {"table": pd.DataFrame({"a": [0], "b": [1]})}
        )

        # Testing that this function executes with no errors
        testbook.ws._write_table_elements([0, 0], gptable, auto_width=True)

    @pytest.mark.parametrize(
        "cell_value1,cell_value2,expectation",
        [
            (None, "valid text", pytest.warns(UserWarning)),
            ("", "valid text", pytest.warns(UserWarning)),
            (" ", "valid text", pytest.warns(UserWarning)),
            ("    ", "valid text", pytest.warns(UserWarning)),
            ("_", "valid text", pytest.raises(ValueError)),
            (" *", "valid text", pytest.raises(ValueError)),
            (" Hello_World! ", "valid text", does_not_raise()),
        ],
    )
    def test__write_table_elements_cell_validation(
        self,
        testbook,
        create_gptable_with_kwargs,
        cell_value1,
        cell_value2,
        expectation,
    ):
        gptable = create_gptable_with_kwargs(
            {
                "table": pd.DataFrame(
                    {
                        "colA": [cell_value1, cell_value2],
                        "colB": ["valid text", "valid text"],
                    }
                )
            }
        )
        with expectation:
            testbook.ws._write_table_elements([0, 0], gptable, auto_width=True)

    def test__apply_column_alignments(self, testbook):
        data_table = pd.DataFrame(
            {
                "index_column": [1, 2],
                "integer_column": [1, 2],
                "float_column": [1.1, 2.2],
                "string_column": ["A", "B"],
                "url_column": [{"display_text": "link"}, {"display_text": "link"}],
                "integer_with_confidential_shorthand": [1, "[c]"],
                "float_with_significant_shorthand": ["1.1[sss]", 2.2],
            }
        )

        format_table = pd.DataFrame(
            {
                "index_column": [{}, {}],
                "integer_column": [{}, {}],
                "float_column": [{}, {}],
                "string_column": [{}, {}],
                "url_column": [{}, {}],
                "integer_with_confidential_shorthand": [{}, {}],
                "float_with_significant_shorthand": [{}, {}],
            }
        )

        testbook.ws._apply_column_alignments(
            data_table, format_table, index_columns=[0]
        )

        exp_format_table = pd.DataFrame(
            {
                "index_column": [{"align": "left"}, {"align": "left"}],
                "integer_column": [{"align": "right"}, {"align": "right"}],
                "float_column": [{"align": "right"}, {"align": "right"}],
                "string_column": [{"align": "left"}, {"align": "left"}],
                "url_column": [{"align": "left"}, {"align": "left"}],
                "integer_with_confidential_shorthand": [
                    {"align": "right"},
                    {"align": "right"},
                ],
                "float_with_significant_shorthand": [
                    {"align": "right"},
                    {"align": "right"},
                ],
            }
        )

        assert_frame_equal(format_table, exp_format_table)


class TestGPWorksheetReferences:
    """
    Test that GPTable note references are modified correctly by GPWorksheet
    during write_gptable().
    """

    @pytest.mark.parametrize("text", test_text_list)
    def test__replace_reference(self, text, testbook):
        """
        Test that references ($$ref$$ style) in strings are replaced with
        [note n], in order of appearance. Also tests replacement in lists.
        """
        got_output = []
        reference_order = ["reference", "another"]

        got_output = [
            testbook.ws._replace_reference(text, reference_order)
            for text in test_text_list
        ]

        exp_refs = ["reference", "another"]
        assert reference_order == exp_refs
        assert got_output == exp_text_list

    @pytest.mark.parametrize(
        "text,refs,output",
        zip(
            test_text_list,
            [["reference"], [], ["another"]],
            ["This has a [note 1]", "This one doesn't", "Here's one[note 2]"],
        ),
    )
    def test__replace_reference_in_attr_str(self, text, refs, output, testbook):
        """
        Test that references are replaced in a single string.
        """
        reference_order = ["reference", "another"]
        got_text = testbook.ws._replace_reference_in_attr(text, reference_order)

        assert got_text == output

    def test__replace_reference_in_attr_dict(self, testbook):
        """
        Test that references are replaced in dictionary values, but not keys.
        """
        reference_order = ["reference", "another"]
        test_text_dict = {
            "$$key$$": "This is a value with a $$reference$$",
            "second_key": "Second value",
            "another_key": "$$another$$reference",
        }
        got_text = testbook.ws._replace_reference_in_attr(
            test_text_dict, reference_order
        )

        exp_text_dict = {
            "$$key$$": "This is a value with a [note 1]",
            "second_key": "Second value",
            "another_key": "reference[note 2]",
        }

        assert got_text == exp_text_dict


class TestGPWorksheetFormatUpdate:
    """
    Test that GPWorksheet format updating methods work as expected.
    """

    def test__apply_format_dict(self, testbook):
        test = dict()
        format_dict = {"bold": True}
        testbook.ws._apply_format(test, format_dict)
        exp = {"bold": True}
        assert test == exp

    def test__apply_format_series(self, testbook):
        test = pd.Series([{} for n in range(3)])
        format_dict = {"bold": True}
        testbook.ws._apply_format(test, format_dict)
        exp = pd.Series([{"bold": True} for n in range(3)])
        assert_series_equal(test, exp)

    def test__apply_format_dataframe(self, testbook):
        test = pd.DataFrame(columns=[0, 1, 2], index=[0, 1])
        test.iloc[0] = [{} for n in range(3)]
        test.iloc[1] = [{} for n in range(3)]

        format_dict = {"bold": True}
        testbook.ws._apply_format(test, format_dict)
        exp = pd.DataFrame(columns=[0, 1, 2], index=[0, 1])
        exp.iloc[0] = [{"bold": True} for n in range(3)]
        exp.iloc[1] = [{"bold": True} for n in range(3)]
        assert_frame_equal(test, exp)


class TestGPWorksheetTable:
    """
    Test that the table property inherited from `xlsxwriter.Worksheet` is set correctly.
    """

    def test__mark_data_as_worksheet_table(self, testbook, create_gptable_with_kwargs):
        df = pd.DataFrame({"col1": ["x", "y"], "col2": [0, 1]})
        gptable = create_gptable_with_kwargs(
            {
                "table": df,
            }
        )
        gptable._set_data_range()

        table_format = pd.DataFrame({"col1": [{}, {}], "col2": [{}, {}]})

        testbook.ws._write_array(
            [0, 2], df, table_format
        )  # First two rows reserved for title and instructions

        testbook.ws._mark_data_as_worksheet_table(gptable, table_format)

        assert len(testbook.ws.tables) == 1

        table = testbook.ws.tables[0]

        got_table_range = table["a_range"]
        exp_table_range = xlsxwriter.utility.xl_range(*gptable.data_range)

        assert got_table_range == exp_table_range

        assert table["name"] == gptable.table_name

        got_number_of_columns = len(table["columns"])
        exp_number_of_columns = df.shape[0]
        assert got_number_of_columns == exp_number_of_columns

        for n in range(got_number_of_columns):
            got_column_name = table["columns"][n]["name"]
            exp_column_name = df.columns[n]
            assert got_column_name == exp_column_name

            got_heading_format = table["columns"][n]["name_format"]
            exp_heading_format = testbook.wb.add_format(table_format.iloc[0, n])
            assert got_heading_format.__dict__ == exp_heading_format.__dict__

    @pytest.mark.parametrize(
        "data,format,exp_width",
        [
            # Single column, normal case
            (["string", "longer string"], [{"font_size": 12}, {"font_size": 12}], [93]),
            # Multiple columns
            (
                pd.DataFrame({"col1": ["a", "bb"], "col2": ["ccc", "dddd"]}),
                pd.DataFrame(
                    {
                        "col1": [{"font_size": 11}, {"font_size": 12}],
                        "col2": [{"font_size": 10}, {"font_size": 14}],
                    }
                ),
                [26, 50],
            ),
            # Bold formatting
            (
                ["bold", "bolder"],
                [{"font_size": 11, "bold": True}, {"font_size": 12, "bold": True}],
                [58],
            ),
            # Multi-line cell
            (
                ["short\nlongest\nmid", "tiny"],
                [{"font_size": 11}, {"font_size": 11}],
                [53],
            ),
            # Empty string
            (["", ""], [{"font_size": 11}, {"font_size": 11}], [0]),
            # Number cell
            ([123, 4567], [{"font_size": 11}, {"font_size": 11}], [35]),
        ],
    )
    def test__calculate_column_widths(self, testbook, data, format, exp_width):
        if isinstance(data, pd.DataFrame):
            table = data
            table_format = format
        else:
            table = pd.DataFrame({"col": data})
            table_format = pd.DataFrame({"col": format})

        got_width = testbook.ws._calculate_column_widths(table, table_format)
        assert got_width == exp_width
        assert all(isinstance(w, int) for w in got_width)

    @pytest.mark.parametrize(
        "format_dict,longest_line,expected",
        [
            ({"font_size": 11, "bold": False}, "abc", 1.0),
            ({"font_size": 12, "bold": False}, "abc", 12 / 11),
            ({"font_size": 11, "bold": True}, "abc", 1.1),
            ({"font_size": 12, "bold": True}, "abc", (12 / 11) * 1.1),
            ({}, "abc", 1.0),
            ({"font_size": 11, "bold": False}, "ABC", 1.0 * (1 + 0.15 * 1)),
            ({"font_size": 11, "bold": False}, "AbC", 1.0 * (1 + 0.15 * (2 / 3))),
            ({"font_size": 11, "bold": True}, "ALLCAPS", 1.1 * (1 + 0.15 * 1)),
            (
                {"font_size": 12, "bold": True},
                "MiXeD",
                (12 / 11) * 1.1 * (1 + 0.15 * (3 / 5)),
            ),
            ({"font_size": 11, "bold": False}, "lower", 1.0),
        ],
    )
    def test__get_scaling_factor(self, testbook, format_dict, longest_line, expected):
        got = testbook.ws._get_scaling_factor(format_dict, longest_line)
        assert got == expected

    @pytest.mark.parametrize(
        "cell_val,expected",
        [
            ("short\nlongest\nmid", "longest"),
            ("one line", "one line"),
            (["a", "bb", "ccc"], "ccc"),
            ("a\nbb\nccc", "ccc"),
        ],
    )
    def test__get_longest_line(self, testbook, cell_val, expected):
        got = testbook.ws._get_longest_line(cell_val)
        assert got == expected

    @pytest.mark.parametrize(
        "cell_val,expected",
        [
            ("abc", "abc"),
            (123, "123"),
            (["a", "b", "c"], "a\nb\nc"),
            ({"x": 1, "y": 2}, "x\ny"),
            (pd.Timestamp("2023-09-30 12:34:56"), "2023-09-30 12:34:56"),
        ],
    )
    def test__get_cell_string(self, testbook, cell_val, expected):
        # Patch FormatList handling if needed
        got = testbook.ws._get_cell_string(cell_val)
        assert got == expected


class TestGPWorkbookStatic:
    """
    Test that the GPWorkbook static methods work as expected.
    """

    @pytest.mark.parametrize(
        "input, expected",
        [
            ("no references", "no references"),
            ("ref at end$$1$$", "ref at end"),
            ("$$1$$ref at start", "ref at start"),
            ("two$$1$$ refs$$2$$", "two refs"),
            ("three$$1$$ refs$$2$$, wow$$3$$", "three refs, wow"),
        ],
    )
    def test__strip_annotation_references(self, input, expected):
        assert GPWorkbook._strip_annotation_references(input) == expected


class TestGPWorkbook:
    """
    Test that GPWorkbook initialisation and methods work as expected.
    """

    def test_subclass(self):
        """
        Test that the GPWorkbook class is a subclass of the XlsxWriter
        Workbook class.
        """
        assert issubclass(GPWorkbook, xlsxwriter.Workbook)

    def test_default_theme_set(self, testbook):
        """
        Test that the workbook theme is set to gptheme by default.
        """
        assert testbook.wb.theme == gptheme

    def test_valid_set_theme(self, testbook):
        """
        Test that setting a new theme with a Theme object works as expected.
        Essentially, make sure that gptheme is not used.
        """
        theme_config = {"title": {"bold": True}}
        theme = gptables.Theme(theme_config)
        testbook.wb.set_theme(theme)

        assert testbook.wb.theme == gptables.Theme(theme_config)

    @pytest.mark.parametrize(
        "not_a_theme", [dict(), set(), [], 1, 3.14, "test_string", pd.DataFrame()]
    )
    def test_invalid_set_theme(self, not_a_theme, testbook):
        """
        Test that setting theme with an object that is not a gptables.Theme
        raises a TypeError.
        """
        with pytest.raises(TypeError):
            testbook.wb.set_theme(not_a_theme)

    def test__update_annotations(self, testbook, create_gptable_with_kwargs):
        """
        Test that _update_annotations produces a correctly ordered list of
        note references used in sheets.
        """
        table = pd.DataFrame(columns=["col"])

        kwargs1 = {
            "title": "Title$$1$$",
            "subtitles": ["Subtitle$$2$$"],
            "units": {0: "Unit$$3$$"},
            "table_notes": {0: "Note$$4$$"},
            "table": table,
        }

        kwargs2 = {
            "title": "Title$$1$$",
            "subtitles": ["Subtitle$$3$$"],
            "units": {0: "Unit$$5$$"},
            "table_notes": {0: "Note$4$$"},
            "table": table,
        }

        gptable1 = create_gptable_with_kwargs(kwargs1)
        gptable2 = create_gptable_with_kwargs(kwargs2)
        sheets = {"sheet1": gptable1, "sheet2": gptable2}

        gpworkbook = testbook.wb
        gpworkbook._update_annotations(sheets)

        assert gpworkbook._annotations == ["1", "2", "3", "4", "5"]

    @pytest.mark.parametrize(
        "additional_elements,values",
        [
            (None, None),
            (["scope"], ["scope"]),
            (
                ["subtitles", "instructions", "scope", "source"],
                [["subtitles"], "instructions", "scope", "source"],
            ),
        ],
    )
    def test_make_table_of_contents(
        self, testbook, create_gptable_with_kwargs, additional_elements, values
    ):
        """
        Test that attributes are set as expected when contentsheet is created.
        """
        kwargs = {}
        if additional_elements:
            kwargs.update(dict(zip(additional_elements, values)))

        exp_toc = pd.DataFrame(
            {
                "Sheet name": [{"sheet": "internal:'sheet'!A1"}],
                "Table description": [["Sheet title", *kwargs.keys()]],
            }
        )
        exp_contentsheet = create_gptable_with_kwargs(
            {
                "table_name": "contents_table",
                "title": "Table of contents",
                "instructions": "This worksheet contains one table.",
                "table": exp_toc,
                "index_columns": {2: 0},
            }
        )

        got_contentsheet = testbook.wb.make_table_of_contents(
            sheets={
                "sheet": create_gptable_with_kwargs({"title": "Sheet title", **kwargs})
            },
            additional_elements=list(kwargs.keys()),
        )

        assert_frame_equal(got_contentsheet.table, exp_contentsheet.table)

        got_contentsheet.table = None
        exp_contentsheet.table = None

        assert got_contentsheet.__dict__ == exp_contentsheet.__dict__

    def test_make_notesheet(self, testbook, create_gptable_with_kwargs):
        """
        Test that creating a notes table sheet using `make_notesheet` generates
        the same gptables.GPTable object as expected.
        """
        gpworkbook = testbook.wb
        gpworkbook._annotations = [1, 2]
        dummy_table = pd.DataFrame(
            data={"Note number": [1, 2], "Note text": ["text", "more text"]}
        )

        notes_name = "Just_a_notesheet"
        notes_title = "Are these the notes you're looking for?"
        notes_instructions = "These are not the notes you're looking for"

        got_notesheet = gpworkbook.make_notesheet(
            notes_table=dummy_table,
            table_name=notes_name,
            title=notes_title,
            instructions=notes_instructions,
        )
        exp_notesheet = create_gptable_with_kwargs(
            {
                "table": dummy_table,
                "table_name": notes_name,
                "title": notes_title,
                "instructions": notes_instructions,
            }
        )

        assert_frame_equal(got_notesheet.table, exp_notesheet.table)

        got_notesheet.table = None
        exp_notesheet.table = None

        assert got_notesheet.__dict__ == exp_notesheet.__dict__

    def test_notesheet_defaults(self, testbook, create_gptable_with_kwargs):
        """
        Test that creating a notes table sheet with arguments set to defaults generates
        the same gptables.GPTable object as expected.
        """
        gpworkbook = testbook.wb
        gpworkbook._annotations = [1, 2]
        dummy_table = pd.DataFrame(
            data={"Note number": [1, 2], "Note text": ["text", "more text"]}
        )

        notes_name = "notes_table"
        notes_title = "Notes"
        notes_instructions = "This worksheet contains one table."

        got_notesheet = gpworkbook.make_notesheet(notes_table=dummy_table)
        exp_notesheet = create_gptable_with_kwargs(
            {
                "table": dummy_table,
                "table_name": notes_name,
                "title": notes_title,
                "instructions": notes_instructions,
            }
        )

        assert_frame_equal(got_notesheet.table, exp_notesheet.table)

        got_notesheet.table = None
        exp_notesheet.table = None

        assert got_notesheet.__dict__ == exp_notesheet.__dict__



# flake8: noqa

from gptables.core.api import (
    produce_workbook,
    write_workbook,
)
from gptables.core.cover import Cover
from gptables.core.gptable import GPTable
from gptables.core.theme import Theme
from gptables.core.wrappers import GPWorkbook
from gptables.utils.unpickle_themes import gptheme

__doc__ = """
*******************************
Good Practice Tables (gptables)
*******************************

.. image:: https://travis-ci.org/best-practice-and-impact/gptables.svg?branch=master
    :target: https://travis-ci.org/best-practice-and-impact/gptables
    :alt: Travis build status

.. image:: https://badge.fury.io/py/gptables.svg
    :target: https://badge.fury.io/py/gptables
    :alt: PyPI release


``gptables`` is an opinionated python package for spreadsheet production.
It produces ``.xlsx`` files from your ``pandas`` dataframes. You define the mapping from your data to elements of the
table. It does the rest.

``gptables`` uses the official `guidance on good practice spreadsheets`_
It advocates a strong adherence to the guidance by restricting the range of
operations possible. The default theme ``gptheme`` should accommodate most use
cases. However, the ``Theme`` object allows development of custom themes, where
other formatting is required.

R users should check out `aftables`_, an R native equivalent to ``gptables``.

``gptables`` is developed and maintained by the `Analysis Function`_.

.. _`guidance on good practice spreadsheets`: https://analysisfunction.civilservice.gov.uk/policy-store/releasing-statistics-in-spreadsheets/

.. _`Analysis Function`: https://analysisfunction.civilservice.gov.uk/



5 Simple Steps
--------------

1. You map your data to the elements of a ``GPTable``.

2. You can define the format of each element with a custom ``Theme``, or simply use the default - gptheme.

3. Optionally design a ``Cover`` page to provide information that relates to all of the tables in your Workbook.

4. Optionally upload a ``notes_table`` with information about any notes.

5. You ``write_workbook`` to win.
"""



"""
Penguins - Wide vs Long Data Example
------------------------------------

This example demonstrates:
- The difference between wide and long data formats
- How to convert between them using pandas
- How to use gptables with both formats

See https://www.statology.org/long-vs-wide-data/ for an
overview of wide vs long data formats.
Also see https://tidyr.tidyverse.org/articles/tidy-data.html
for more on tidy data principles.
"""

from pathlib import Path

import pandas as pd

import gptables as gpt

# Read data
parent_dir = Path(__file__).parents[1]
penguins_data = pd.read_csv(parent_dir / "test/data/penguins.csv")

# --- Wide Format ---
# The original penguins_data is in wide format: each measurement is a separate column

wide_table = gpt.GPTable(
    table=penguins_data,
    table_name="penguins_wide",
    title="Penguins Dataset (Wide Format)",
    subtitles=[
        "Each measurement is a separate column.",
        "Wide format is common for spreadsheets and some analyses.",
    ],
    scope="Penguins",
    source="Palmer Station, Antarctica",
)

# --- Long Format ---
# Convert to long format using melt
# This example combines measurements into a single column
long_df = pd.melt(
    penguins_data,
    id_vars=["Species", "Island", "Sex"],
    value_vars=[
        "Culmen Length (mm)",
        "Culmen Depth (mm)",
        "Flipper Length (mm)",
        "Body Mass (g)",
    ],
    var_name="Measurement",
    value_name="Value",
)

long_table = gpt.GPTable(
    table=long_df,
    table_name="penguins_long",
    title="Penguins Dataset (Long Format)",
    subtitles=[
        "Measurements are stacked in a single column.",
        "Long format is preferred for tidy data and many analyses.",
    ],
    scope="Penguins",
    source="Palmer Station, Antarctica",
)

if __name__ == "__main__":
    output_path = parent_dir / "python_penguins_wide_long_gptable.xlsx"
    gpt.write_workbook(
        filename=output_path,
        sheets={"Wide Format": wide_table, "Long Format": long_table},
        contentsheet_options={"additional_elements": ["subtitles", "scope"]},
    )
    print("Output written at: ", output_path)



from functools import wraps
from typing import Any, Callable, Dict, List, Optional, Union

import yaml
from xlsxwriter.format import Format


def validate_single_format(f: Callable) -> Callable:
    @wraps(f)
    def wrapper(cls, format_dict: Dict[str, Any]) -> Any:
        """
        Decorator to validate that input is a dictionary dictionary.
        """
        if not isinstance(format_dict, dict):
            actual_type = type(format_dict)
            msg = "Formats must be supplied as a dictionary, not" f"{actual_type}"
            raise ValueError(msg)

        for fmt in format_dict.keys():
            cls._validate_format_label(fmt)
        return f(cls, format_dict)

    return wrapper


class Theme:
    """
    Defines format attributes for table elements for use with XlsxWriter.

    For valid values, see XlsxWriter `Format` properties:
    _<https://xlsxwriter.readthedocs.io/format.html#format-methods-and-format-properties>_.

    Attributes
    ----------
    cover_title_format : dict

    cover_subtitle_format : dict

    cover_text_format : dict

    title_format : dict

    subtitle_format : dict

    instructions_format : dict

    scope_format : dict

    column_heading_format : dict

    index_1_format : dict

    index_2_format : dict

    index_3_format : dict

    data_format : dict

    source_format : dict

    legend_format : dict

    description_order : list
    """

    def __init__(self, config: Optional[Union[Dict[str, Any], str]] = None) -> None:
        """
        Initialise theme object.

        Parameters
        ----------
        config : dict or .yaml/.yml file
            theme specification
        """
        # Formats
        self._format_attributes = [
            "cover_title_format",
            "cover_subtitle_format",
            "cover_text_format",
            "title_format",
            "subtitle_format",
            "instructions_format",
            "scope_format",
            "column_heading_format",
            "index_1_format",
            "index_2_format",
            "index_3_format",
            "data_format",
            "source_format",
            "legend_format",
        ]

        for attr in self._format_attributes:
            setattr(self, attr, {})

        # Other attributes
        self.description_order = []

        # Valid Them format attributes
        self._valid_attrs = [
            x.replace("_format", "") for x in self._format_attributes
        ] + ["global"]

        # Valid XlsxWriter Format attributes
        self._valid_format_labels = [
            attr.replace("set_", "")
            for attr in Format().__dir__()
            if attr.startswith("set_") and callable(getattr(Format(), attr))
        ]

        if config:
            self.apply_config(config)

    @staticmethod
    def _parse_config(config: Union[Dict[str, Any], str]) -> Dict[str, Any]:
        """
        Parse yaml configuration to dictionary.
        """
        if isinstance(config, str):
            if not config.endswith((".yml", ".yaml")):
                raise ValueError("Theme configuration files must be YAML")
            with open(config, "r") as file:
                cfg = yaml.safe_load(file)

        elif isinstance(config, dict):
            cfg = config

        else:
            raise ValueError("Theme configuration must be a dict or YAML file")

        return cfg

    def _validate_config(self, config: Dict[str, Any]) -> None:
        """
        Assert that format dictionary lower level keys are valid XlsxWriter
        Format attributes.
        """
        for attr in config.keys():
            if attr in self._valid_attrs:
                attr_config = config[attr] or {}
                for fmt in attr_config.keys():
                    self._validate_format_label(fmt)

    def _validate_format_label(self, format_name: str) -> None:
        """
        Assert that format is a valid XlsxWriter Format attribute.
        """
        if format_name not in self._valid_format_labels:
            raise ValueError(f"`{format_name}` is not a valid format label")

    def apply_config(self, config: Union[Dict[str, Any], str]) -> None:
        """
        Update multiple Theme attributes using a YAML or dictionary config.
        This enables extension of build in Themes.
        """
        cfg = self._parse_config(config)
        self._validate_config(cfg)

        # Update all when global used
        if "global" in cfg.keys():
            default_format = cfg.pop("global")
            self._update_all_formats(default_format)

        # Update with individual methods
        for key, value in cfg.items():
            if key == "description_order":
                getattr(self, "update_" + key)(value)
            elif key in self._valid_attrs:
                if value is not None:
                    getattr(self, "update_" + key + "_format")(value)
            else:
                raise ValueError(f"`{key}` is not a valid Theme attribute")

    def _update_all_formats(self, global_dict: Dict[str, Any]) -> None:
        """
        Updates all theme attributes with a global format dictionary.
        """
        for attr in self._format_attributes:
            if attr.endswith("_format"):
                getattr(self, "update_" + attr)(global_dict)

    @validate_single_format
    def update_column_heading_format(self, format_dict: Dict[str, Any]) -> None:
        """
        Update the `column_heading_format` attribute. Where keys already exist,
        existing items are replaced.
        """
        self.column_heading_format.update(format_dict)

    @validate_single_format
    def update_index_1_format(self, format_dict: Dict[str, Any]) -> None:
        """
        Update the `index_1_format` attribute. Where keys already exist, existing
        items are replaced.
        """
        self.index_1_format.update(format_dict)

    @validate_single_format
    def update_index_2_format(self, format_dict: Dict[str, Any]) -> None:
        """
        Update the `index_2_format` attribute. Where keys already exist, existing
        items are replaced.
        """
        self.index_2_format.update(format_dict)

    @validate_single_format
    def update_index_3_format(self, format_dict: Dict[str, Any]) -> None:
        """
        Update the `index_3_format` attribute. Where keys already exist, existing
        items are replaced.
        """
        self.index_3_format.update(format_dict)

    @validate_single_format
    def update_data_format(self, format_dict: Dict[str, Any]) -> None:
        """
        Update the `data_format` attribute. Where keys already exist,
        existing items are replaced.
        """
        self.data_format.update(format_dict)

    @validate_single_format
    def update_cover_title_format(self, format_dict: Dict[str, Any]) -> None:
        """
        Update the `cover_title_format` attribute. Where keys already exist, existing
        items are replaced.
        """
        self.cover_title_format.update(format_dict)

    @validate_single_format
    def update_cover_subtitle_format(self, format_dict: Dict[str, Any]) -> None:
        """
        Update the `cover_subtitle_format` attribute. Where keys already exist, existing
        items are replaced.
        """
        self.cover_subtitle_format.update(format_dict)

    @validate_single_format
    def update_cover_text_format(self, format_dict: Dict[str, Any]) -> None:
        """
        Update the `cover_text_format` attribute. Where keys already exist, existing
        items are replaced.
        """
        self.cover_text_format.update(format_dict)

    @validate_single_format
    def update_title_format(self, format_dict: Dict[str, Any]) -> None:
        """
        Update the `title_format` attribute. Where keys already exist, existing
        items are replaced.
        """
        self.title_format.update(format_dict)

    @validate_single_format
    def update_subtitle_format(self, format_dict: Dict[str, Any]) -> None:
        """
        Update the `subtitle_format` attribute. Where keys already exist,
        existing items are replaced.
        """
        self.subtitle_format.update(format_dict)

    @validate_single_format
    def update_instructions_format(self, format_dict: Dict[str, Any]) -> None:
        """
        Update the `instructions_format` attribute. Where keys already exist,
        existing items are replaced.
        """
        self.instructions_format.update(format_dict)

    @validate_single_format
    def update_scope_format(self, format_dict: Dict[str, Any]) -> None:
        """
        Update the `scope_format` attribute. Where keys already exist, existing
        items are replaced.
        """
        self.scope_format.update(format_dict)

    @validate_single_format
    def update_location_format(self, format_dict: Dict[str, Any]) -> None:
        """
        Update the `location_format` attribute. Where keys already exist,
        existing items are replaced.
        """
        self.location_format.update(format_dict)

    @validate_single_format
    def update_source_format(self, format_dict: Dict[str, Any]) -> None:
        """
        Update the `source_format` attribute. Where keys already exist,
        existing items are replaced.
        """
        self.source_format.update(format_dict)

    @validate_single_format
    def update_legend_format(self, format_dict: Dict[str, Any]) -> None:
        """
        Update the `legend_format` attribute. Where keys already exist,
        existing items are replaced.
        """
        self.legend_format.update(format_dict)

    def update_description_order(self, order_list: List[str]) -> None:
        """
        Update the `description_order` attribute. Overrides existing order.
        """
        if not isinstance(order_list, list):
            msg = "`description_order` must be a list of description element names"
            raise TypeError(msg)

        valid_elements = ["instructions", "source", "legend", "scope"]
        if not all(element in valid_elements for element in order_list):
            msg = f"`description_order` elements must be in {valid_elements}"
            raise ValueError(msg)
        self.description_order = order_list

    def print_attributes(self) -> None:
        """
        Print all current format attributes and values to the console.
        """
        obj_attr = [
            attr
            for attr in self.__dir__()
            if not attr.startswith("_") and not callable(getattr(self, attr))
        ]
        for attr in obj_attr:
            print(attr, ":", getattr(self, attr))

    def __eq__(self, other: object) -> bool:
        """
        Comparison operator, for testing.
        """
        # don't attempt to compare against unrelated types
        if not isinstance(other, Theme):
            return False

        obj_attr = [
            attr
            for attr in self.__dir__()
            if not attr.startswith("_") and not callable(getattr(self, attr))
        ]
        return all([getattr(self, attr) == getattr(other, attr) for attr in obj_attr])



from contextlib import contextmanager

import pandas as pd
import pytest
from pandas.testing import assert_frame_equal

# TODO: These should be stored in GPTable
gptable_compulsory_text_attrs = ["title", "instructions"]

gptable_optional_text_attrs = ["scope", "source"]

gptable_list_text_attrs = ["subtitles", "legend"]


valid_index_columns = [
    {},
    {1: "one"},
    {1: "one", 2: "two"},
    {1: "one", 2: "two", 3: "three"},
]

valid_text_elements_excl_none = [
    "This is a string",
    ["This is ", {"bold": True}, "rich", "text"],
]

valid_text_elements_incl_none = valid_text_elements_excl_none.copy()
valid_text_elements_incl_none.append(None)

invalid_text_elements_excl_none = [dict(), set(), 42, 3.14, True]

invalid_text_elements_incl_none = invalid_text_elements_excl_none.copy()
invalid_text_elements_incl_none.append(None)


@contextmanager
def does_not_raise():
    yield


def test_init_defaults(create_gptable_with_kwargs):
    """
    Test that given a minimal input, default attributes are correct types.
    """
    empty_gptable = create_gptable_with_kwargs()

    # Required args
    assert empty_gptable.title == ""
    assert empty_gptable.table_name == "table_name"

    assert_frame_equal(empty_gptable.table, pd.DataFrame().reset_index(drop=True))

    # Optional args
    assert empty_gptable.scope is None
    assert empty_gptable.source is None
    assert empty_gptable.units is None
    assert empty_gptable.index_columns == {}
    assert empty_gptable.subtitles == []
    assert empty_gptable.legend == []
    assert empty_gptable.additional_formatting == []
    assert (
        empty_gptable.instructions
        == "This worksheet contains one table. Some cells may refer to notes, which can be found on the notes worksheet."
    )

    # Other
    assert empty_gptable.index_levels == 0
    assert empty_gptable._column_headings == set()
    assert empty_gptable._VALID_INDEX_LEVELS == [1, 2, 3]
    assert empty_gptable._annotations == []
    assert empty_gptable.data_range == [2, 0, 2, -1]


class TestAttrValidationGPTable:
    @pytest.mark.parametrize("level", [4, 0, -1])
    def test_invalid_index_level(self, level, create_gptable_with_kwargs):
        """
        Test that GPTable index_columns raises error when an index level is
        invalid.
        """
        with pytest.raises(ValueError):
            create_gptable_with_kwargs(
                {
                    "table": pd.DataFrame(columns=["one", "two", "three"]),
                    # Valid column index, but invalid level
                    "index_columns": {level: 0},
                }
            )

    @pytest.mark.parametrize(
        "idx,expectation",
        [
            (0, does_not_raise()),
            (1, does_not_raise()),
            (3, pytest.raises(ValueError)),
            (-1, pytest.raises(ValueError)),
        ],
    )
    def test_invalid_column_index(self, idx, expectation, create_gptable_with_kwargs):
        """
        Test that GPTable index_columns raises error when a column index number
        is invalid.
        """
        with expectation:
            create_gptable_with_kwargs(
                {
                    "table": pd.DataFrame(columns=["one", "two", "three"]),
                    # Valid index level, but non existent column
                    "index_columns": {1: idx},
                }
            )

    @pytest.mark.parametrize(
        "index_cols,col_headings",
        zip(valid_index_columns, [{0, 1, 2, 3}, {1, 2, 3}, {2, 3}, {3}]),
    )
    def test_set_column_index(
        self, index_cols, col_headings, create_gptable_with_kwargs
    ):
        """
        Test that setting GPTable index_columns with valid column index works
        as expected.
        """
        gptable = create_gptable_with_kwargs(
            {
                "table": pd.DataFrame(columns=["one", "two", "three", "four"]),
                "index_columns": index_cols,
            }
        )
        assert gptable.index_columns == index_cols
        assert gptable.index_levels == len(index_cols)
        assert gptable._column_headings == col_headings

    @pytest.mark.parametrize("not_a_table", [1, "?", [1, 3, 5, 6]])
    def test_invalid_table_type(self, not_a_table, create_gptable_with_kwargs):
        """
        Test that setting GPTable table to object that is not a
        pandas.DataFrame raises an error.
        """
        with pytest.raises(TypeError):
            create_gptable_with_kwargs({"table": not_a_table})

    def test_set_table_name(self, create_gptable_with_kwargs):
        """
        Test that setting GPTable table name with a valid string works as expected
        """
        gptable = create_gptable_with_kwargs({"table_name": "table_name"})
        assert gptable.table_name == "table_name"

    @pytest.mark.parametrize("invalid_name", invalid_text_elements_incl_none)
    def test_invalid_type_table_name(self, invalid_name, create_gptable_with_kwargs):
        """
        Test that setting GPTable table name to object other than string raises
        an error.
        """
        with pytest.raises(TypeError):
            create_gptable_with_kwargs({"table_name": invalid_name})

    def test_table_name_not_list(self, create_gptable_with_kwargs):
        """
        Test that setting GPTable table name to a list, eg with rich text,
        raises an error.
        """
        with pytest.raises(TypeError):
            create_gptable_with_kwargs({"table_name": []})

    @pytest.mark.parametrize("invalid_name", [" ", "  "])
    def test_invalid_characters_table_name(
        self, invalid_name, create_gptable_with_kwargs
    ):
        """
        Test that setting GPTable table name to string with whitespace raises
        an error.
        """
        with pytest.raises(ValueError):
            create_gptable_with_kwargs({"table_name": invalid_name})

    @pytest.mark.parametrize("attr", gptable_compulsory_text_attrs)
    @pytest.mark.parametrize("not_text", invalid_text_elements_incl_none)
    def test_invalid_text_in_compulsory_str_attrs(
        self, attr, not_text, create_gptable_with_kwargs
    ):
        """
        Test that setting an invalid GPTable text types raises a TypeError for
        each attribute that holds a string.
        """
        with pytest.raises(TypeError):
            create_gptable_with_kwargs({attr: not_text})

    @pytest.mark.parametrize("attr", gptable_optional_text_attrs)
    @pytest.mark.parametrize("not_text", invalid_text_elements_excl_none)
    def test_invalid_text_in_optional_str_attrs(
        self, attr, not_text, create_gptable_with_kwargs
    ):
        """
        Test that setting an invalid GPTable text types raises a TypeError for
        each attribute that holds a string.
        """
        with pytest.raises(TypeError):
            create_gptable_with_kwargs({attr: not_text})

    @pytest.mark.parametrize("attr", gptable_compulsory_text_attrs)
    @pytest.mark.parametrize("text", valid_text_elements_excl_none)
    def test_valid_text_in_compulsory_str_attrs(
        self, attr, text, create_gptable_with_kwargs
    ):
        """
        Test that setting valid GPTable text elements works as expected. Test
        strings and list containing strings and format dicts (rich text).
        Also test that None is allowed.
        """
        gptable = create_gptable_with_kwargs({attr: text})

        if isinstance(text, list):
            assert getattr(gptable, attr).list == text
        else:
            assert getattr(gptable, attr) == text

    @pytest.mark.parametrize("attr", gptable_optional_text_attrs)
    @pytest.mark.parametrize("text", valid_text_elements_incl_none)
    def test_valid_text_in_optional_str_attrs(
        self, attr, text, create_gptable_with_kwargs
    ):
        """
        Test that setting valid GPTable text elements works as expected. Test
        strings and list containing strings and format dicts (rich text).
        Also test that None is allowed.
        """
        gptable = create_gptable_with_kwargs({attr: text})

        if isinstance(text, list):
            assert getattr(gptable, attr).list == text
        else:
            assert getattr(gptable, attr) == text

    @pytest.mark.parametrize("attr", gptable_list_text_attrs)
    @pytest.mark.parametrize("text", invalid_text_elements_incl_none)
    def test_invalid_text_in_list_attrs(self, attr, text, create_gptable_with_kwargs):
        """
        Test that setting list of invalid text elements to GPTable list
        parameters raises a TypeError.
        """
        text = [text, text]
        with pytest.raises(TypeError):
            create_gptable_with_kwargs({attr: text})

    @pytest.mark.parametrize("attr", gptable_list_text_attrs)
    @pytest.mark.parametrize("text", valid_text_elements_excl_none)
    def test_valid_text_in_list_attrs(self, attr, text, create_gptable_with_kwargs):
        """
        Test that setting list of valid text elements to GPTable list
        parameters works as expected.
        """
        text_list = [text, text]
        gptable = create_gptable_with_kwargs({attr: text_list})

        if isinstance(text, list):
            assert all([element.list == text for element in getattr(gptable, attr)])
        else:
            assert getattr(gptable, attr) == text_list

    @pytest.mark.parametrize(
        "key", invalid_text_elements_incl_none[2:] + ["invalid_key"]
    )
    def test_invalid_additional_format_keys(self, key, create_gptable_with_kwargs):
        """
        Test that adding additional formatting with an invalid key raises an
        error.
        """
        with pytest.raises(ValueError):
            create_gptable_with_kwargs(
                {"additional_formatting": [{key: {"format": {"bold": True}}}]}
            )

    @pytest.mark.parametrize("key", ["cell", "row", "column"])
    def test_valid_additional_format_keys(self, key, create_gptable_with_kwargs):
        """
        Test that adding additional formatting with a valid key (column, row or
        cell) works as expected.
        """
        gptable = create_gptable_with_kwargs(
            {"additional_formatting": [{key: {"format": {"bold": True}}}]}
        )
        assert getattr(gptable, "additional_formatting") == [
            {key: {"format": {"bold": True}}}
        ]

    @pytest.mark.parametrize(
        "format_label", invalid_text_elements_incl_none[2:] + ["not_a_format"]
    )
    def test_invalid_additional_format_labels(
        self, format_label, create_gptable_with_kwargs
    ):
        """
        Test that adding additional formatting with a format parameter that is
        not supported by XlsxWriter raises an error.
        """
        with pytest.raises(ValueError):
            create_gptable_with_kwargs(
                {
                    "additional_formatting": [
                        {"cell": {"format": {format_label: True}, "cells": (0, 0)}}
                    ]
                }
            )

    @pytest.mark.parametrize(
        "format_dict",
        [
            {"bold": True},
            {"font_size": 17},
            {"align": "center"},
            {"font_color": "red"},
            {"bottom": 1},
        ],
    )
    def test_valid_additional_format_labels(
        self, format_dict, create_gptable_with_kwargs
    ):
        """
        Test that adding additional formatting with a format parameter that is
         supported by XlsxWriter works as expected.
        """
        additional_formatting = [{"cell": {"format": format_dict, "cells": (0, 0)}}]
        gptable = create_gptable_with_kwargs(
            {"additional_formatting": additional_formatting}
        )
        assert getattr(gptable, "additional_formatting") == additional_formatting

    @pytest.mark.parametrize("unit_text", valid_text_elements_excl_none)
    @pytest.mark.parametrize("column_id", ["columnA", 0])
    def test_units_placement(self, unit_text, column_id, create_gptable_with_kwargs):
        """
        Test that units are placed correctly under column headers.
        """
        table_with_units = pd.DataFrame(columns=[f"columnA\n({unit_text})"])

        gptable = create_gptable_with_kwargs(
            {
                "table": pd.DataFrame(columns=["columnA"]),
            }
        )

        gptable.set_units(new_units={column_id: unit_text})

        assert gptable.table.columns == table_with_units.columns

    @pytest.mark.parametrize("column_id", ["columnA", 0])
    def test_table_notes_placement(self, column_id, create_gptable_with_kwargs):
        """
        Test that notes are placed correctly under column headers.
        """
        expected_columns = pd.Index(["columnA"])

        gptable = create_gptable_with_kwargs(
            {"table": pd.DataFrame(columns=["columnA"])}
        )

        gptable.set_table_notes(new_table_notes={column_id: "$$note_reference$$"})

        assert gptable.table.columns.equals(expected_columns)

        note_key = column_id if isinstance(column_id, int) else "columnA"
        assert gptable.table_notes == {note_key: "$$note_reference$$"}

        assert all("$$" not in str(h) for h in gptable.table.columns)

    @pytest.mark.parametrize("column_id", ["columnA", 0])
    def test_table_units_and_notes_placement(
        self, column_id, create_gptable_with_kwargs
    ):
        """
        Test that units and notes are placed correctly under column headers.
        """
        expected_columns = pd.Index(["columnA\n(unit)"])

        gptable = create_gptable_with_kwargs(
            {"table": pd.DataFrame(columns=["columnA"])}
        )

        gptable.set_units(new_units={column_id: "unit"})
        gptable.set_table_notes(new_table_notes={column_id: "$$note_reference$$"})

        assert gptable.table.columns.equals(expected_columns)
        note_key = column_id if isinstance(column_id, int) else "columnA"
        assert gptable.table_notes == {note_key: "$$note_reference$$"}
        assert all("$$" not in str(h) for h in gptable.table.columns)

    def test_additional_formatting_with_units(self, create_gptable_with_kwargs):
        """
        Test that units are placed correctly under column headers.
        """

        gptable = create_gptable_with_kwargs(
            {
                "table": pd.DataFrame(columns=["columnA"]),
                "units": {"columnA": "unit"},
                "additional_formatting": [
                    {"column": {"columns": ["columnA"], "format": {"bold": True}}}
                ],
            }
        )

        assert gptable.additional_formatting == [
            {"column": {"columns": ["columnA\n(unit)"], "format": {"bold": True}}}
        ]

    def test_additional_formatting_with_table_notes(self, create_gptable_with_kwargs):
        """
        Test that units are placed correctly under column headers.
        """

        gptable = create_gptable_with_kwargs(
            {
                "table": pd.DataFrame(columns=["columnA"]),
                "table_notes": {"columnA": "$$ref$$"},
                "additional_formatting": [
                    {"column": {"columns": ["columnA"], "format": {"bold": True}}}
                ],
            }
        )

        assert gptable.additional_formatting == [
            {"column": {"columns": ["columnA"], "format": {"bold": True}}}
        ]
        assert gptable.table_notes == {"columnA": "$$ref$$"}
        assert all("$$" not in str(h) for h in gptable.table.columns)

    def test_additional_formatting_with_units_and_table_notes(
        self, create_gptable_with_kwargs
    ):
        """
        Test that units are placed correctly under column headers.
        """

        gptable = create_gptable_with_kwargs(
            {
                "table": pd.DataFrame(columns=["columnA"]),
                "units": {0: "unit"},
                "table_notes": {0: "$$ref$$"},
                "additional_formatting": [
                    {"column": {"columns": ["columnA"], "format": {"bold": True}}}
                ],
            }
        )

        assert gptable.additional_formatting == [
            {
                "column": {
                    "columns": ["columnA\n(unit)"],
                    "format": {"bold": True},
                }
            }
        ]

        assert gptable.units == {0: "unit"}
        assert gptable.table_notes == {0: "$$ref$$"}

    @pytest.mark.parametrize(
        "column_names,expectation",
        [
            (["columnA", "columnB"], does_not_raise()),
            (["columnA", ""], pytest.raises(ValueError)),
            ([None, "columnB"], pytest.raises(ValueError)),
        ],
    )
    def test__validate_all_column_names_have_text(
        self, column_names, expectation, create_gptable_with_kwargs
    ):
        """
        Test that GPTable raises error when there are null values or empty strings for column names.
        """
        with expectation:
            create_gptable_with_kwargs({"table": pd.DataFrame(columns=column_names)})

    @pytest.mark.parametrize(
        "column_names,expectation",
        [
            (["columnA", "columnB", "columnC"], does_not_raise()),
            (["columnA", "columnB", "columnB"], pytest.raises(ValueError)),
        ],
    )
    def test__validate_no_duplicate_column_names(
        self, column_names, expectation, create_gptable_with_kwargs
    ):
        """
        Test that GPTable raises error when there are duplicate column names in table data.
        """
        with expectation:
            create_gptable_with_kwargs({"table": pd.DataFrame(columns=column_names)})


class TestIndirectAttrs:
    """
    Test that non-formatting create_gptable_with_kwargs attributes are indirectly set correctly.
    """

    @pytest.mark.parametrize("index_cols", valid_index_columns)
    def test_index_levels_set(self, index_cols, create_gptable_with_kwargs):
        """
        Test that number of index levels are set, when one, two or three
        indexes are used.
        """
        gptable = create_gptable_with_kwargs(
            {
                "table": pd.DataFrame(columns=["one", "two", "three", "four"]),
                "index_columns": index_cols,
            }
        )
        assert getattr(gptable, "index_levels") == len(index_cols)

    @pytest.mark.parametrize("index_cols", valid_index_columns)
    def test_column_headings_set(self, index_cols, create_gptable_with_kwargs):
        """
        Test that non-index columns are set as column headings.
        """
        gptable = create_gptable_with_kwargs(
            {
                "table": pd.DataFrame(columns=["one", "two", "three", "four"]),
                "index_columns": index_cols,
            }
        )

        # Expect all column numbers that have no index level assigned
        exp = set(range(4)) - set(range(len(index_cols)))

        assert getattr(gptable, "_column_headings") == exp

    def test__annotations_set(self, create_gptable_with_kwargs):
        """
        Test that annotation references in `gptable` attributes are found and
        added to `_annotations` as expected.
        """
        table = pd.DataFrame(columns=["col"])

        kwargs = {
            "title": "Title$$1$$",
            "subtitles": ["Subtitle$$2$$"],
            "instructions": "Instructions$$3$$",
            "source": "Source$$4$$",
            "scope": "Scope$$5$$",
            "legend": ["Legend$$6$$"],
            "units": {0: "Unit$$7$$"},
            "table_notes": {0: "Note$$8$$"},
            "table": table,
        }

        gptable = create_gptable_with_kwargs(kwargs)

        description_order = ["instructions", "source", "scope", "legend"]
        gptable._set_annotations(description_order)

        assert gptable._annotations == ["1", "2", "3", "4", "5", "6", "7", "8"]



from pathlib import Path

import pandas as pd

import gptables as gpt

parent_dir = Path(__file__).parents[1]
penguins_data = pd.read_csv(parent_dir / "test/data/penguins.csv")

penguins_table = gpt.GPTable(
    table=penguins_data,
    table_name="penguins_statistics",
    title="The Palmer Penguins Dataset",
    subtitles=["This is the first subtitle", "This is another subtitle"],
    scope="Penguins",
    source="Palmer Station, Antarctica",
)

penguins_sheets = {"Penguins": penguins_table}

if __name__ == "__main__":
    output_path = parent_dir / "gpt_custom_theme.xlsx"
    theme_path = str(Path(__file__).parent.parent / "themes/example_theme_basic.yaml")
    gpt.write_workbook(
        filename=output_path,
        sheets=penguins_sheets,
        theme=gpt.Theme(theme_path),
        contentsheet_options={"additional_elements": ["subtitles", "scope"]},
    )
    print("Output written at: ", output_path)



"""
Penguins - End-to-End Example
--------------------------------------------

This example demonstrates a full workflow:
- Data loading
- Cleaning (handling missing values, renaming columns, recoding values)
- Rounding measurement columns
- Converting between wide and long formats
- Preparing data for gptables
- Applying additional formatting

"""

from pathlib import Path

import pandas as pd

import gptables as gpt

# Read data
parent_dir = Path(__file__).parents[1]
penguins_data = pd.read_csv(parent_dir / "test/data/penguins.csv")

# --- Data Cleaning ---
# Drop rows with missing values in columns other than "Comments"
cols_to_check = [col for col in penguins_data.columns if col != "Comments"]
cleaned = penguins_data.dropna(subset=cols_to_check)

# Rename columns
cleaned = cleaned.rename(
    columns={
        "Culmen Length (mm)": "Bill Length (mm)",
        "Culmen Depth (mm)": "Bill Depth (mm)",
    }
)

# Recode Sex column
cleaned["Sex"] = cleaned["Sex"].replace({"MALE": "M", "FEMALE": "F"})

# --- Rounding ---
# Round measurement columns to nearest integer
measurement_cols = [
    "Bill Length (mm)",
    "Bill Depth (mm)",
    "Flipper Length (mm)",
    "Body Mass (g)",
]
for col in measurement_cols:
    if col in cleaned.columns:
        cleaned[col] = cleaned[col].round(0).astype("Int64")

# --- Wide to Long Conversion ---
# Example: melt measurements into a long format
long_df = pd.melt(
    cleaned,
    id_vars=["Species", "Island", "Sex"],
    value_vars=[
        "Bill Length (mm)",
        "Bill Depth (mm)",
        "Flipper Length (mm)",
        "Body Mass (g)",
    ],
    var_name="Measurement",
    value_name="Value",
)
# --- Prepare for gptables ---
table_name = "penguins_long_format"
title = "Penguins Dataset (Long Format Example)"
subtitles = ["Demonstrates data cleaning, rounding, and wide-to-long conversion."]
scope = "Penguins"
source = "Palmer Station, Antarctica"

# Additional formatting: highlight 'Value' column
additional_formatting = [
    {"column": {"columns": ["Value"], "format": {"bg_color": "#DDEEFF"}}}
]

penguins_table = gpt.GPTable(
    table=long_df,
    table_name=table_name,
    title=title,
    subtitles=subtitles,
    scope=scope,
    source=source,
    additional_formatting=additional_formatting,
)

if __name__ == "__main__":
    output_path = parent_dir / "python_penguins_end_to_end_gptable.xlsx"
    gpt.write_workbook(
        filename=output_path,
        sheets={"Penguins (Long Format)": penguins_table},
        contentsheet_options={"additional_elements": ["subtitles", "scope"]},
    )
    print("Output written at: ", output_path)



import re
from typing import Any, Dict, List, Optional, Union

import pandas as pd
from xlsxwriter.format import Format


class GPTable:
    """
    A Good Practice Table.

    Stores data, text content, and content metadata for writing a table to Excel.

    Attributes
    ----------
    table : pandas.DataFrame
        table to be written to an Excel workbook
    table_name : str
        name for table. Should be unique with no spaces and always begin with a
        letter, an underscore character, or a backslash. Use letters, numbers,
        periods, and underscore characters for the rest of the name.
    title : str
        title of the table
    subtitles : List[str], optional
        subtitles as a list of strings
    instructions : str, optional
        instructions on how to read the sheet. If not provided, defaults to
        "This worksheet contains one table. Some cells may refer to notes,
        which can be found on the notes worksheet."
    scope : str, optional
        description of scope/basis of data in table if not included in title
    source : str, optional
        description of the source of the data in table if not included in cover
    units : dict, optional
        units used in each (dict) column of table
    legend : list, optional
        descriptions of special notation used in table
    index_columns : dict, optional
        mapping an index level to a 0-indexed column as {level: column}.
        Default is a level two index in the first column ({2: 0}).
    additional_formatting : dict, optional
        table-specific formatting for columns, rows or individual cells
    """

    def __init__(
        self,
        table: pd.DataFrame,
        table_name: str,
        title: str,
        scope: Optional[str] = None,
        source: Optional[str] = None,
        units: Optional[Dict[Any, Any]] = None,
        table_notes: Optional[Dict[Any, Any]] = None,
        subtitles: Optional[List[Any]] = [],
        instructions: str = "",
        legend: Optional[List[Any]] = [],
        index_columns: Optional[Dict[int, int]] = None,
        additional_formatting: Optional[List[Dict[str, Any]]] = [],
    ) -> None:

        # Attributes
        self.title = None
        self.subtitles = []

        self.units = None  # str or {units (str):column index (int)} dict
        self.table_notes = None  # str or {units (str):column index (int)} dict

        self._VALID_INDEX_LEVELS = [1, 2, 3]
        self.index_levels = 0
        self.index_columns = {}  # {index level (int): column index (int)}
        self._column_headings = set()  # Non-index column headings
        self.table = pd.DataFrame()
        self.table_name = None
        self.data_range = [0] * 4

        self.scope = None
        self.source = None
        self.legend = []
        self._annotations = []

        self.additional_formatting = []

        if additional_formatting is None:
            additional_formatting = []

        self.additional_formatting = additional_formatting

        # Valid format labels from XlsxWriter
        self._valid_format_labels = [
            attr.replace("set_", "")
            for attr in Format().__dir__()
            if attr.startswith("set_") and callable(getattr(Format(), attr))
        ]

        # Call methods to set attributes
        self.set_title(title)
        self.set_subtitles(subtitles)
        self.set_instructions(instructions)
        self.set_additional_formatting(additional_formatting)
        self.set_table(table, index_columns, units, table_notes)
        self.set_table_name(table_name)
        self.set_scope(scope)
        self.set_source(source)
        self.set_legend(legend)
        self._set_data_range()

    def set_table(
        self,
        new_table: pd.DataFrame,
        new_index_columns: Optional[Dict[int, int]] = None,
        new_units: Optional[Dict[Any, Any]] = None,
        new_table_notes: Optional[Dict[Any, Any]] = None,
    ) -> None:
        """
        Set the `table`, `index_columns`, `units` and `table_notes` attributes. Overwrites
        existing values for these attributes.
        """
        if not isinstance(new_table, pd.DataFrame):
            raise TypeError("`table` must be a pandas DataFrame")

        if any("$$" in str(h) for h in new_table.columns):
            raise ValueError(
                "Notes inside column headers are no longer supported. "
                "Use GPTable.table_notes for column notes instead."
            )

        default_index = pd.Index(range(new_table.shape[0]))
        if not all(new_table.index == default_index) and not new_table.empty:
            msg = (
                "`table` index must not contain index data. It can be reset"
                " before adding to a GPTable (see DataFrame.reset_index())."
                " Please ensure that index data is stored in the first 1-3"
                " columns of `table` and is indicated in `index_columns`."
            )
            raise ValueError(msg)

        self.table = new_table.reset_index(drop=True)

        self._validate_all_column_names_have_text()
        self._validate_no_duplicate_column_names()

        if new_index_columns is None:
            new_index_columns = self.index_columns
        self.set_index_columns(new_index_columns)

        if new_units is None:
            new_units = self.units
        self.set_units(new_units)

        if new_table_notes is None:
            new_table_notes = self.table_notes
        self.table_notes = new_table_notes

    def set_index_columns(self, new_index_columns: Dict[int, int]) -> None:
        """
        Set the `index_columns` attribute. Overwrites any existing values.
        A dict must be supplied. This dict should map index level to a
        single 0-indexed column number. All other columns will be considered
        as data columns.
        """
        if isinstance(new_index_columns, dict):
            # Check if levels and values are valid
            valid_levels = all(
                level in self._VALID_INDEX_LEVELS for level in new_index_columns.keys()
            )
            if not valid_levels:
                msg = (
                    "`index_columns` dictionary keys must be valid index"
                    f" levels: {self._VALID_INDEX_LEVELS}"
                )
                raise ValueError(msg)

            if not all(isinstance(col, int) for col in new_index_columns.values()):
                # Convert col name to numeric index
                for key, value in new_index_columns.items():
                    col_iloc = self.table.columns.get_loc(value)
                    new_index_columns.update({key: col_iloc})

            column_indexes = [col for col in new_index_columns.values()]

            valid_columns = all(self._valid_column_index(col) for col in column_indexes)
            if not valid_columns:
                msg = (
                    "Out of range - `index_columns` dictionary values must"
                    "be valid, 0-indexed column numbers"
                )
                raise ValueError(msg)

            self.index_levels = len(new_index_columns.keys())

            self.index_columns = new_index_columns
            self._set_column_headings()
        else:
            msg = (
                "`index_columns` must be a dict mapping a valid index level"
                " to a 0-indexed column number"
            )
            raise ValueError(msg)

    def _valid_column_index(self, column_index: int) -> bool:
        """
        Check if `column_index` is valid, given the `table` shape.
        """
        return column_index in range(self.table.shape[1])

    def _set_column_headings(self) -> None:  # TODO: check custom formatting in headers
        """
        Sets the `column_headings` attribute to the set of column indexes that
        are not assigned to `index_columns`.
        """
        index_cols = set(self.index_columns.values())
        self._column_headings = {x for x in range(self.table.shape[1])} - index_cols

    def _validate_all_column_names_have_text(self) -> None:
        """
        Validate that all column names in header row have text.
        """
        for column_name in self.table.columns:
            if pd.isna(column_name):
                msg = "Null column name found in table data - column names must all have text"
                raise ValueError(msg)
            elif len(column_name) > 0:
                continue
            else:
                msg = "Empty column name found in table data - column names must all have text"
                raise ValueError(msg)

    def _validate_no_duplicate_column_names(self) -> None:
        """
        Validate that there are no duplicate column names in table data.
        """
        if len(self.table.columns) != len(set(self.table.columns)):
            msg = "Duplicate column names found in table data - column names must be unique"
            raise ValueError(msg)

    def set_table_name(self, new_table_name: str) -> None:
        """
        Set the `table_name` attribute.
        """
        if not isinstance(new_table_name, str):
            msg = "table_name should be provided as a string"
            raise TypeError(msg)

        elif len(new_table_name) != len("".join(new_table_name.split())):
            msg = "Whitespace found in table_name, remove or replace with underscores"
            raise ValueError(msg)

        else:
            self.table_name = new_table_name

    def set_title(self, new_title: Any) -> None:
        """
        Set the `title` attribute.
        """
        self._validate_text(new_title, "title")

        if isinstance(new_title, list):
            new_title = FormatList(new_title)

        self.title = new_title

    def add_subtitle(self, new_subtitle: Any) -> None:
        """
        Add a single subtitle to the existing list of `subtitles`.
        """
        self._validate_text(new_subtitle, "subtitles")

        if isinstance(new_subtitle, list):
            new_subtitle = FormatList(new_subtitle)

        self.subtitles.append(new_subtitle)

    def set_subtitles(
        self, new_subtitles: Optional[List[Any]], overwrite: bool = True
    ) -> None:
        """
        Set a list of subtitles to the `subtitles` attribute. Overwrites
        existing ist of subtitles by default. If `overwrite` is False, new list
        is appended to existing list of subtitles.
        """
        if new_subtitles is None:
            new_subtitles = []

        if not isinstance(new_subtitles, (list)):
            msg = (
                "`subtitles` must be provided as a list containing strings"
                " and/or lists of strings and format dictionaries"
                " (rich text)"
            )
            raise TypeError(msg)

        for text in new_subtitles:
            self._validate_text(text, "subtitles")

        new_subtitles = [
            FormatList(text) if isinstance(text, list) else text
            for text in new_subtitles
        ]

        if overwrite:
            self.subtitles = new_subtitles
        else:
            self.subtitles += new_subtitles

    def set_instructions(self, new_instructions: Any) -> None:
        """
        Set `instructions` attribute.
        """
        self._validate_text(new_instructions, "instructions")

        if len(new_instructions) == 0:
            self.instructions = "This worksheet contains one table. Some cells may refer to notes, which can be found on the notes worksheet."
        elif isinstance(new_instructions, list):
            self.instructions = FormatList(new_instructions)
        else:
            self.instructions = new_instructions

    def set_scope(self, new_scope: Any) -> None:
        """
        Set the `scope` attribute.
        """
        if new_scope is None:
            new_scope = ""
            return

        self._validate_text(new_scope, "scope")

        if isinstance(new_scope, list):
            new_scope = FormatList(new_scope)

        self.scope = new_scope

    def set_units(
        self, new_units: Optional[Dict[Any, Any]]
    ) -> None:  # TODO: custom formatting in units?
        """
        Adds units to column headers.
        Units should be in the format {column: units_text}. Column can be column name or 0-indexed column
        number in `table`.
        """
        if any("$$" in str(h) for h in self.table.columns):
            raise ValueError(
                "Notes inside column headers are no longer supported. "
                "Use GPTable.table_notes for column notes instead."
            )
        if isinstance(new_units, dict) and len(new_units) > 0:
            for value in new_units.values():
                self._validate_text(value, "units")

            headers = self.table.columns.values.tolist()

            # Check if notes have already been added to headers...
            unmodified_headers = [header.split("\n")[0] for header in headers]

            # ...if so, apply any units applied to headers without notes, to headers with notes
            for n in range(len(unmodified_headers)):
                if unmodified_headers[n] in list(new_units.keys()):
                    new_units[n] = new_units.pop(unmodified_headers[n])

            # Convert numeric keys to column names
            new_headers_keys = [
                headers[key] if isinstance(key, int) else key
                for key in new_units.keys()
            ]
            new_headers_values = [
                f"{key}\n({value})"
                for key, value in zip(new_headers_keys, new_units.values())
            ]
            new_headers = dict(zip(new_headers_keys, new_headers_values))

            self.table = self.table.rename(columns=new_headers)

            if len(self.additional_formatting) > 0:
                self._update_column_names_in_additional_formatting(new_headers)

        elif new_units is not None:
            msg = (
                "`units` attribute must be a dictionary or None"
                " ({column: units_text})"
            )

            raise TypeError(msg)

        self.units = new_units

    def _update_column_names_in_additional_formatting(
        self, col_names: Dict[Any, Any]
    ) -> None:
        """
        Parameters
        ----------
        col_names: dict
            with keys old names and values new names, where new names are old names plus units
        Return
        ------
        None
        """
        formatting_list = self.additional_formatting
        for dictionary in formatting_list:
            if list(dictionary.keys()) == ["column"]:
                format = list(dictionary.values())[0]

                # new_name if name==old_name else name for name in col_names
                format["columns"] = [
                    col_names[name] if name in list(col_names.keys()) else name
                    for name in format["columns"]
                ]

        self.additional_formatting = formatting_list

    def set_table_notes(
        self, new_table_notes: Optional[Dict[Any, Any]]
    ) -> None:  # TODO: custom formatting in column headers?
        """
        Adds note references to column headers.
        `table_notes` should be in the format {column: "$$note_reference$$"}.
        Column can be column name or 0-indexed column number in `table`.

        """
        self.table_notes = new_table_notes

    def set_source(self, new_source: Any) -> None:
        """
        Set the source attribute to the specified str.
        """
        if new_source is None:
            new_source = ""
            return

        self._validate_text(new_source, "source")

        if isinstance(new_source, list):
            new_source = FormatList(new_source)

        self.source = new_source

    def add_legend(self, new_legend: Any) -> None:
        """
        Add a single legend entry to the existing `legend` list.
        """
        self._validate_text(new_legend, "legend")

        if isinstance(new_legend, list):
            new_legend = FormatList(new_legend)

        self.legend.append(new_legend)

    def set_legend(
        self, new_legend: Optional[List[Any]], overwrite: bool = True
    ) -> None:
        """
        Set a list of legend entries to the `legend` attribute. Overwrites
        existing legend entries by default. If overwrite is False, new entries
        are appended to the `legend` list.
        """
        if new_legend is None:
            self.legend = []
            return
        if not isinstance(new_legend, list):
            msg = "`legend` must be provided as a list of text elements"
            raise TypeError(msg)
        for text in new_legend:
            self._validate_text(text, "legend")

        new_legend = [
            FormatList(text) if isinstance(text, list) else text for text in new_legend
        ]

        if overwrite:
            self.legend = new_legend
        else:
            self.legend += new_legend

    def _set_annotations(self, description_order: List[str]) -> None:
        """
        Set a list of note references to the `_annotations` attribute.
        """
        elements = [
            "title",
            "subtitles",
            *description_order,
            "units",
            "table_notes",
        ]

        ordered_refs = []

        for attr in elements:
            attr_current = getattr(self, attr)
            references = self._get_references_from_attr(attr_current)
            ordered_refs.extend(references)

        # Deprecated as of v1.1.0 - instead use `table_notes` to add references to column headers
        table_refs = self._get_references_from_table()
        ordered_refs.extend(table_refs)

        # remove duplicates from ordered_refs and assign to self._annotations
        self._annotations = list(dict.fromkeys(ordered_refs))

    def _get_references_from_attr(self, data: Any) -> List[str]:
        """
        Finds references in a string or list/dict of strings. Works
        recursively on list elements and dict values. Other types are ignored.
        Returns ordered list of references from attribute.

        Parameters
        ----------
        data : string or list/dict of strings
            object containing strings to replace references in

        Returns
        -------
        list of str
        """
        ordered_refs = []
        if isinstance(data, str):
            ordered_refs.extend(self._get_references(data))
        if isinstance(data, list):
            for n in range(len(data)):
                ordered_refs.extend(self._get_references_from_attr(data[n]))
        if isinstance(data, dict):
            for key in data.keys():
                ordered_refs.extend(self._get_references(data[key]))
        if isinstance(data, FormatList):
            data_list = data.list
            for n in range(len(data_list)):
                if isinstance(data_list[n], str):
                    ordered_refs.extend(self._get_references(data_list[n]))

        return ordered_refs

    # Deprecated as of v1.1.0 - instead use `table_notes` to add references to column headers
    def _get_references_from_table(self) -> List[str]:
        """
        Get note references in the table column headings and index columns.
        """
        table = self.table

        ordered_refs = []
        column_references = self._get_references_from_attr(table.columns.to_list())
        ordered_refs.extend(column_references)

        index_columns = self.index_columns.values()
        for col in index_columns:
            index_column = table.iloc[:, col]
            index_column_references = self._get_references_from_attr(
                index_column.to_list()
            )
            ordered_refs.extend(index_column_references)

        return ordered_refs

    @staticmethod
    def _get_references(string: str) -> List[str]:
        """
        Given a single string, return occurrences of note references (denoted by
        flanking dollar signs [$$reference$$]).

        Parameters
        ----------
        string : str
            the string to find references within

        Returns
        -------
        list of str
            list of note references
        """
        ordered_refs = []
        refs_raw = re.findall(r"[$]{2}.*?[$]{2}", string)
        refs_clean = [w.replace("$", "") for w in refs_raw]
        ordered_refs.extend(refs_clean)

        return ordered_refs

    def set_additional_formatting(self, new_formatting: List[Dict[str, Any]]) -> None:
        """
        Set a dictionary of additional formatting to be applied to this table.
        """
        if not isinstance(new_formatting, list):
            msg = "`additional_formatting` must be a list of dictionaries"
            raise TypeError(msg)
        keys = [key for item in new_formatting for key in item.keys()]
        for key in keys:
            if key not in ["column", "row", "cell"]:
                msg = (
                    f"`{key}` is not a supported format type. Please use"
                    " `column`, `row` or `cell`"
                )
                raise ValueError(msg)

        self._validate_format_labels(new_formatting)

        self.additional_formatting = new_formatting

    def _validate_format_labels(self, format_list: List[Dict[str, Any]]) -> None:
        """
        Validate that format labels are valid property of XlsxWriter Format.
        """
        labels = [
            label
            for item in format_list
            for key in item.keys()
            for label in item[key]["format"]
        ]
        for label in labels:
            if label not in self._valid_format_labels:
                msg = f"`{label}` is not a valid XlsxWriter Format property"
                raise ValueError(msg)

    def _set_data_range(self) -> None:
        """
        Get the top-left and bottom-right cell reference of the table data.
        """
        # TODO: ugly code
        row_offset = (
            sum(
                [
                    int(self.title is not None),
                    int(self.scope is not None),
                    int(self.source is not None),
                ]
            )
            + 1
        )  # corresponds to instructions which are included by default

        if self.subtitles is not None:
            row_offset += len(self.subtitles)
        if self.legend is not None:
            row_offset += len(self.legend)

        self.data_range = [
            row_offset,
            0,
            self.table.shape[0] + row_offset,
            self.table.shape[1] - 1,
        ]

    @staticmethod
    def _validate_text(obj: Any, attr: str) -> None:
        """
        Validate that an object contains valid text elements. These are either
        strings or list of strings and dictionaries.
        """
        if isinstance(obj, str):
            return None

        if isinstance(obj, list):
            for element in obj:
                if not isinstance(element, (str, dict)):
                    msg = (
                        f"{attr} text should be provided as strings or"
                        " lists of strings and dictionaries (rich-text)."
                        f" {type(element)} are not valid rich text"
                        " elements."
                    )
                    raise TypeError(msg)
        else:
            msg = (
                f"{attr} text should be provided as strings or lists of"
                f" strings and dictionaries (rich-text). {type(obj)} are"
                " not valid text elements."
            )
            raise TypeError(msg)


class FormatList:
    """
    Class for storing list of alternating string and dictionary objects.
    Dictionaries specify additional formatting to be applied to the following string.
    """

    def __init__(self, items: List[Union[str, Dict[str, Any]]]) -> None:
        self.list = items
        self._set_string_property()

    def _set_string_property(self) -> None:
        string = ""
        for entry in self.list:
            if isinstance(entry, str):
                string += entry

        self.string = string



import pickle  # nosec B403

import importlib_resources


class ThemeUnpickler(pickle.Unpickler):
    """
    Points the unpickler to the Theme class. Allows unpickling for package
    init.
    """

    def find_class(self, module: str, name: str) -> type:
        if name == "Theme":
            from gptables.core.theme import Theme

            return Theme
        return super().find_class(module, name)


file = importlib_resources.files("gptables") / "theme_pickles/gptheme.pickle"

with importlib_resources.as_file(file) as path:
    gptheme = ThemeUnpickler(open(path, "rb")).load()



from pathlib import Path

import pandas as pd

import gptables as gpt

parent_dir = Path(__file__).parents[1]
penguins_data = pd.read_csv(parent_dir / "test/data/penguins.csv")

penguins_table = gpt.GPTable(
    table=penguins_data,
    table_name="penguins_statistics",
    title="The Palmer Penguins Dataset$$note_about_x$$",
    subtitles=[
        "This is the first subtitle$$note_about_y$$",
        "This is another subtitle",
    ],
    scope="Penguins",
    source="Palmer Station, Antarctica",
)

penguins_sheets = {"Penguins": penguins_table}

notes = {
    "Note reference": [
        "note_about_x",
        "note_about_y",
        "note_about_z",
        "note_with_no_link",
    ],
    "Note text": [
        "This is a note about x linking to google.",
        "This is a note about y linking to duckduckgo.",
        "This is a note about z linking to the ONS website.",
        "This is a note with no link.",
    ],
    "Useful link": [
        "[google](https://www.google.com)",
        "[duckduckgo](https://duckduckgo.com/)",
        "[ONS](https://www.ons.gov.uk)",
        None,
    ],
}
penguins_notes_table = pd.DataFrame.from_dict(notes)

if __name__ == "__main__":
    output_path = parent_dir / "gpt_adding_notes.xlsx"
    gpt.write_workbook(
        filename=output_path,
        sheets=penguins_sheets,
        notes_table=penguins_notes_table,
        contentsheet_options={"additional_elements": ["subtitles", "scope"]},
    )
    print("Output written at: ", output_path)



import re
import warnings
from copy import deepcopy
from math import ceil

import numpy as np
import pandas as pd
from xlsxwriter.utility import cell_autofit_width
from xlsxwriter.workbook import Workbook
from xlsxwriter.worksheet import Worksheet

from gptables.core.cover import Cover
from gptables.utils.unpickle_themes import gptheme

from .gptable import FormatList, GPTable
from .theme import Theme


class GPWorksheet(Worksheet):
    """
    Wrapper for an `XlsxWriter.Worksheet` object. Provides a method for writing
    a good practice table (GPTable) to a Worksheet.
    """

    def write_cover(self, cover: "Cover") -> None:
        """
        Write a cover page to the Worksheet. Uses text from a Cover object and
        details of the Workbook contents.

        Parameters
        ----------
        cover : gptables.Cover
            object containing cover sheet text
        """
        theme = self.theme
        pos = [0, 0]

        self._parse_urls(cover)

        pos = self._write_element(pos, cover.title, theme.cover_title_format)

        if cover.intro is not None:
            pos = self._write_element(
                pos, "Introductory information", theme.cover_subtitle_format
            )
            pos = self._write_element_list(pos, cover.intro, theme.cover_text_format)

        if cover.about is not None:
            pos = self._write_element(
                pos, "About these data", theme.cover_subtitle_format
            )
            pos = self._write_element_list(pos, cover.about, theme.cover_text_format)

        if cover.contact is not None:
            pos = self._write_element(pos, "Contact", theme.cover_subtitle_format)
            pos = self._write_element_list(pos, cover.contact, theme.cover_text_format)

        self.set_column(0, 0, cover.width)

    def write_gptable(
        self,
        gptable: "GPTable",
        auto_width: bool,
        reference_order: list = [],
    ) -> None:
        """
        Write data from a GPTable object to the worksheet using the workbook
        Theme object for formatting.

        Parameters
        ----------
        gptable : gptables.GPTable
            object containing elements of the gptable to be written to the
            Worksheet
        reference_order : list, optional
            order of annotations in workbook
            must be provided if gptable uses annotations
        Returns
        -------
        None
        """
        if not isinstance(gptable, GPTable):
            raise TypeError("`gptable` must be a gptables.GPTable object")

        if len(gptable._annotations) > 0 and len(reference_order) == 0:
            msg = "reference_order must be provided if gptable contains annotations"
            raise ValueError(msg)

        theme = self.theme

        # Write each GPTable element using appropriate Theme attr
        pos = [0, 0]

        self._reference_annotations(gptable, reference_order)
        self._parse_urls(gptable)

        gptable = deepcopy(gptable)

        pos = self._write_element(pos, gptable.title, theme.title_format)

        pos = self._write_element_list(pos, gptable.subtitles, theme.subtitle_format)

        description = theme.description_order
        for element in description:
            pos = getattr(self, "_write_" + element)(
                pos, getattr(gptable, element), getattr(theme, element + "_format")
            )

        pos = self._write_table_elements(
            pos,
            gptable,
            auto_width,
        )

    def _reference_annotations(
        self,
        gptable: "GPTable",
        reference_order: list,
    ) -> None:
        """
        Replace note references with numbered references and move to end of element.
        Acts on `title`, `subtitles`, `table` and `notes` attributes of a GPTable.
        References are numbered from top left of spreadsheet, working across each row.

        Parameters
        ----------
        gptable : gptables.GPTable
            object containing data with references to notes
        reference_order : list
            order of annotations in workbook

        Returns
        -------
        None
        """
        description_order = self.theme.description_order

        elements = [
            "title",
            "subtitles",
            *description_order,
        ]

        # Loop through elements, replacing references in strings
        for attr in elements:
            attr_current = getattr(gptable, attr)
            setattr(
                gptable,
                attr,
                self._replace_reference_in_attr(attr_current, reference_order),
            )
        self._reference_table_annotations(gptable, reference_order)

    def _reference_table_annotations(
        self,
        gptable: "GPTable",
        reference_order: list,
    ) -> None:
        """
        Reference annotations in the table column headings and index columns.
        """
        table = gptable.table.copy()

        notes = getattr(gptable, "table_notes", {}) or {}
        if notes:
            headers = list(table.columns)
            rename_map = {}

            for key, note_token in notes.items():

                idx = key if isinstance(key, int) else table.columns.get_loc(key)
                old = headers[idx]

                rendered = self._replace_reference_in_attr(note_token, reference_order)

                if not (
                    isinstance(old, str)
                    and old.splitlines()
                    and old.splitlines()[-1] == rendered
                ):
                    new = f"{old}\n{rendered}"
                    rename_map[old] = new
            if rename_map:
                table = table.rename(columns=rename_map)

        index_columns = gptable.index_columns.values()
        for col in index_columns:
            table.iloc[:, col] = table.iloc[:, col].apply(
                lambda x: self._replace_reference_in_attr(x, reference_order)
            )
        gptable.table = table

    def _replace_reference_in_attr(
        self,
        data: object,
        reference_order: list,
    ) -> object:
        """
        Replaces references in a string or list/dict of strings. Works
        recursively on list elements and dict values. Other types are returned
        without modification.

        Parameters
        ----------
        data : any type
            object containing strings to replace references in
        reference_order : list
            order of annotations in workbook

        Returns
        -------
        string : str
            input string with references replaced with numerical reference (n),
            where n is the order of appearance in the resulting document
        """
        if isinstance(data, str):
            data = self._replace_reference(data, reference_order)
        if isinstance(data, list):
            for n in range(len(data)):
                data[n] = self._replace_reference_in_attr(data[n], reference_order)
        if isinstance(data, dict):
            for key in data.keys():
                data[key] = self._replace_reference_in_attr(data[key], reference_order)
        if isinstance(data, FormatList):
            data_list = data.list
            for n in range(len(data_list)):
                data_list[n] = self._replace_reference_in_attr(
                    data_list[n], reference_order
                )
            data = FormatList(data_list)

        return data

    @staticmethod
    def _replace_reference(string: str, reference_order: list) -> str:
        """
        Given a single string, record occurrences of new references (denoted by
        flanking dollar signs [$$reference$$]) and replace with number
        reference reflecting order of detection.

        Parameters
        ----------
        string : str
            the string to replace references within
        reference_order : list
            order of annotations in workbook

        Returns
        -------
        string : str
            input string with references replaced with numerical reference (n),
            where n is the order of appearence in the resulting document
        """
        text_refs = re.findall(r"[$]{2}.*?[$]{2}", string)
        dict_refs = [w.replace("$", "") for w in text_refs]
        for n in range(len(dict_refs)):
            num_ref = "[note " + str(reference_order.index(dict_refs[n]) + 1) + "]"
            string = string.replace(text_refs[n], "") + num_ref

        return string

    def _parse_urls(self, sheet: object) -> None:
        """
        Convert markdown URL formatting into URL, string tuple

        Parameters
        ----------
        sheet : gptables.GPTable, gptables.Cover
            object containing data with urls
        """
        if isinstance(sheet, GPTable):
            elements = [
                "title",
                "subtitles",
                "legend",
                "source",
                "scope",
                "units",
            ]
        elif isinstance(sheet, Cover):
            elements = [
                "title",
                "intro",
                "about",
                "contact",
            ]

        # Loop through elements, replacing urls in strings
        for attr in elements:
            attr_current = getattr(sheet, attr)
            setattr(
                sheet,
                attr,
                self._replace_url_in_attr(
                    attr_current,
                ),
            )
        if isinstance(sheet, GPTable):
            self._parse_table_urls(sheet)

    def _parse_table_urls(self, gptable: "GPTable") -> None:
        """
        Parse URLs in table.
        """
        table = getattr(gptable, "table")
        rows, columns = table.shape

        for c in range(columns):
            for r in range(rows):
                cell = self._replace_url_in_attr(table.iloc[r, c])
                if isinstance(cell, dict):
                    table.iloc[r, c] = [cell]
                else:
                    table.iloc[r, c] = cell

        setattr(gptable, "table", table)

    def _replace_url_in_attr(self, data: object) -> object:
        """
        Replaces urls in a string or list/dict of strings. Works
        recursively on list elements and dict values. Other types
        are returned without modification.

        Parameters
        ----------
        data : any type
            object containing strings to replace references in
        """
        if isinstance(data, str):
            data = self._replace_url(data)
        if isinstance(data, list):
            for n in range(len(data)):
                data[n] = self._replace_url_in_attr(
                    data[n],
                )
        if isinstance(data, dict):
            for key in data.keys():
                data[key] = self._replace_url_in_attr(
                    data[key],
                )

        return data

    @staticmethod
    def _replace_url(string: str) -> object:
        """
        Given a single string, record occurrences of markdown
        style urls (formatted as `"[url](display_text)"`) and
        replace with tuples of `(url, string)`

        Parameters
        ----------
        string : str
            the string to replace references within

        Returns
        -------
        string or dict
            if no markdown style urls found, returns sting
            if found, return dictionary with key `string` and value `url`,
            where markdown style url in `string` is replaced with `display_text`
        """
        f_url_pattern = r"\[.+\]\(.+\)"  # "[display_text](url)"
        f_urls = re.findall(f_url_pattern, string)

        if len(f_urls) == 0:
            return string

        if len(f_urls) > 1:
            msg = (
                "More than one link found in cell. Excel only permits one link per cell"
            )
            raise ValueError(msg)
        else:
            f_url = f_urls[0]

            url = re.split(r"\(", f_url)[1].replace(")", "")
            display_text = re.split(r"\]", f_url)[0].replace("[", "")

            string = re.sub(f_url_pattern, display_text, string)

            return {string: url}

    def _write_element(self, pos: list, element: object, format_dict: dict) -> list:
        """
        Write a single text element of a GPTable to the GPWorksheet.

        Parameters
        ----------
        element : str or list
            the string or list of rich string elements to be written
        format_dict : dict
            format to be applied to string
        pos : list
            the position of the worksheet cell to write the element to

        Returns
        -------
        pos : list
            new position to write next element from
        """
        if element:
            self._smart_write(*pos, element, format_dict)
            pos[0] += 1

        return pos

    def _write_element_list(
        self, pos: list, element_list: list, format_dict: dict
    ) -> list:
        """
        Writes a list of elements row-wise.

        Parameters
        ----------
        element_list : list
            list of strings or nested list of rich string elements to write,
            one per row
        format_dict : dict
            format to be applied to string
        pos : list
            the position of the worksheet cell to write the elements to

        Returns
        -------
        pos: list
            new position to write next element from
        """
        if element_list:
            for element in element_list:
                pos = self._write_element(pos, element, format_dict)

        return pos

    def _write_instructions(
        self, pos: list, element: object, format_dict: dict
    ) -> list:
        """
        Alias for writting description elements by name.
        """
        return self._write_element(pos, element, format_dict)

    def _write_source(self, pos: list, element: object, format_dict: dict) -> list:
        """
        Alias for writting description elements by name.
        """
        return self._write_element(pos, element, format_dict)

    def _write_scope(self, pos: list, element: object, format_dict: dict) -> list:
        """
        Alias for writting description elements by name.
        """
        return self._write_element(pos, element, format_dict)

    def _write_legend(self, pos: list, element_list: list, format_dict: dict) -> list:
        """
        Alias for writting description elements by name.
        """
        return self._write_element_list(pos, element_list, format_dict)

    def _write_notes(self, pos: list, element_list: list, format_dict: dict) -> list:
        """
        Alias for writting description elements by name.
        """
        return self._write_element_list(pos, element_list, format_dict)

    def _write_table_elements(
        self, pos: list, gptable: "GPTable", auto_width: bool
    ) -> list:
        """
        Writes the table and units elements of a GPTable. Uses the
        Workbook Theme, plus any additional formatting associated with the
        GPTable.

        Parameters
        ----------
        gptable : gptables.GPTable
            object containing the table and additional formatting data
        pos : list
            the position of the worksheet cell to write the units to
        auto_width : bool
            select if column widths should be determined automatically using
            length of text in index and columns

        Returns
        -------
        pos : list
            new position to write next element from
        """
        # Convert whitespace only cells to None
        gptable.table.replace({r"^\s*$": None}, inplace=True, regex=True)

        if gptable.table.isna().values.all():
            msg = f"""
            {gptable.table_name} contains only null or whitespace cells.
            Please provide alternative table containing data.
            """
            raise ValueError(msg)

        if gptable.table.isna().all(axis=1).any():
            msg = f"""
            Empty or null row found in {gptable.table_name}.
            Please remove blank rows before passing data to GPTable.
            """
            raise ValueError(msg)

        if gptable.table.isna().values.any():
            msg = f"""
            Empty or null cell found in {gptable.table_name}. The reason for
            missingness should be included in the `GPTable.instructions` attribute.
            There should only be one reason otherwise a shorthand should be
            provided in the `instructions` or `legend` attribute.
            Guidance on shorthand can be found at:
            https://analysisfunction.civilservice.gov.uk/policy-store/symbols-in-tables-definitions-and-help/
            """
            warnings.warn(msg)

        # Raise error if any table element is only special characters
        if gptable.table.astype("string").stack().str.contains("^[^a-zA-Z0-9]*$").any():
            msg = f"""
            Cell found in {gptable.table_name} containing only special characters,
            replace with alphanumeric characters before inputting to GPTable.
            Guidance on symbols in tables can be found at:
            https://analysisfunction.civilservice.gov.uk/policy-store/symbols-in-tables-definitions-and-help/
            """
            raise ValueError(msg)

        # Get theme
        theme = self.theme

        # Reset position to left col on next row
        pos[1] = 0

        # Create data array
        index_levels = gptable.index_levels
        index_columns = [col for col in gptable.index_columns.values()]
        data = pd.DataFrame(gptable.table, copy=True)

        # Create row containing column headings
        data.loc[-1] = data.columns
        data.index = data.index + 1
        data.sort_index(inplace=True)

        # Create formats array
        # pandas.DataFrame did NOT want to hold dictionaries, so be wary
        formats = pd.DataFrame().reindex_like(data).astype(object)
        dict_row = [{} for n in range(formats.shape[1])]
        for row in range(formats.shape[0]):
            dict_row = [{} for n in range(formats.shape[1])]
            formats.iloc[row] = dict_row

        # Add Theme formatting to formats dataframe
        format_headings_from = 0
        self._apply_format(
            formats.iloc[0, format_headings_from:], theme.column_heading_format
        )

        self._apply_format(formats.iloc[1:, index_levels:], theme.data_format)

        index_level_formats = [
            theme.index_1_format,
            theme.index_2_format,
            theme.index_3_format,
        ]
        for level, col in gptable.index_columns.items():
            self._apply_format(
                formats.iloc[1:, col],
                index_level_formats[level - 1],  # Account for 0-indexing
            )

        self._apply_column_alignments(data, formats, index_columns)

        # Add additional table-specific formatting from GPTable
        self._apply_additional_formatting(
            formats, gptable.additional_formatting, gptable.index_levels
        )

        # Write table
        pos = self._write_array(pos, data, formats)

        # Set columns widths
        if auto_width:
            widths = self._calculate_column_widths(data, formats)
            self._set_column_widths(widths)

        self._mark_data_as_worksheet_table(gptable, formats)

        return pos

    def _apply_column_alignments(
        self, data_table: pd.DataFrame, formats_table: pd.DataFrame, index_columns: list
    ) -> None:
        """
        Add column alignment to format based on datatype

        Parameters
        ----------
        data_table : pandas.DataFrame
            table to be written to an Excel workbook
        formats_table : pandas.DataFrame
            table with same dimensions as `data_table`,
            containing formating dictionaries

        """
        # look for shorthand notation, usually a few letters in square brackets
        # will also find note markers eg [Note 1]
        # Using np.nan instead on None for backwards compatibility with pandas <=1.4
        with pd.option_context("future.no_silent_downcasting", True):
            data_table_copy = data_table.replace(
                regex=r"\[[\w\s]+\]",
                value=np.nan,
            ).infer_objects(copy=False)
            if (
                data_table_copy.columns.to_list()
                == data_table_copy.iloc[0, :].to_list()
            ):
                # drop first row which contains column names
                data_table_copy = data_table_copy.iloc[1:]

        data_table_copy = data_table_copy.convert_dtypes()

        column_types = data_table_copy.dtypes
        if index_columns == [] and pd.api.types.is_numeric_dtype(
            column_types[data_table.columns[0]]
        ):
            column_types[data_table.columns[0]] = "str"

        for column in data_table.columns:
            if data_table.columns.get_loc(column) in index_columns:
                alignment_dict = {"align": "left"}

            elif pd.api.types.is_numeric_dtype(column_types[column]):
                alignment_dict = {"align": "right"}

            else:
                alignment_dict = {"align": "left"}

            self._apply_format(formats_table[column], alignment_dict)

    def _apply_additional_formatting(
        self,
        formats_table: pd.DataFrame,
        additional_formatting: list,
        index_levels: int,
    ) -> None:
        """
        Apply row, column and cell formatting to dataframe of formats.
        """
        for item in additional_formatting:
            fmt_type = list(item.keys())[0]
            format_desc = item[fmt_type]

            if fmt_type == "cell":
                formatting = format_desc["format"]
                cell_ilocs = format_desc["cells"]
                if isinstance(cell_ilocs, tuple):
                    cell_ilocs = [cell_ilocs]
                for row, col in cell_ilocs:
                    formats_table_slice = formats_table.iloc[row, col]

            elif fmt_type == "column":
                cols_iloc = [
                    formats_table.columns.get_loc(col) if isinstance(col, str) else col
                    for col in format_desc["columns"]
                ]
                row_start = 0
                if "include_names" in format_desc.keys():
                    row_start = 0 if format_desc["include_names"] else 1

                formats_table_slice = formats_table.iloc[row_start:, cols_iloc]
                formatting = format_desc["format"]

            elif fmt_type == "row":
                rows_iloc = format_desc["rows"]
                col_start = 0
                if "include_names" in format_desc.keys():
                    col_start = 0 if format_desc["include_names"] else index_levels

                formats_table_slice = formats_table.iloc[rows_iloc, col_start:]
                formatting = format_desc["format"]

            self._apply_format(formats_table_slice, formatting)

    def _write_array(
        self, pos: list, data: pd.DataFrame, formats: pd.DataFrame
    ) -> list:
        """
        Write a two-dimensional array to the current Worksheet, starting from
        the specified position.

        Parameters
        ----------
        data : pandas.DataFrame
            array of data to be written to Worksheet
        formats : pandas.DataFrame
            array of dictionaries that specify the formatting to be applied
            to each cell of data
        pos : list
            the position of the top left cell to start writing the array from

        Returns
        -------
        pos : list
            new position to write next element from
        """
        if data.shape != formats.shape:
            raise ValueError("data and formats arrays must be of equal shape")

        rows, cols = data.shape
        for row in range(rows):
            for col in range(cols):
                cell_data = data.iloc[row, col]
                cell_format_dict = formats.iloc[row, col]

                self._smart_write(
                    pos[0] + row, pos[1] + col, cell_data, cell_format_dict
                )

        pos = [pos[0] + rows, 0]

        return pos

    def _mark_data_as_worksheet_table(
        self, gptable: "GPTable", formats_dataframe: pd.DataFrame
    ) -> None:
        """
        Marks the data to be recognised as a Worksheet Table in Excel.

        Parameters
        ----------
        gptable : gptables.GPTable
            object containing the table
        formats_dataframe : DataFrame
            DataFrame with same dimensions as gptable.table, containing
            formatting dictionaries
        """
        data_range = gptable.data_range

        column_list = gptable.table.columns.tolist()
        formats_list = [
            self._workbook.add_format(format_dict)
            for format_dict in formats_dataframe.iloc[0, :].tolist()
        ]

        column_headers = [
            {"header": header, "header_format": header_format}
            for header, header_format in zip(column_list, formats_list)
        ]

        self.add_table(
            *data_range,
            {
                "header_row": True,
                "autofilter": False,
                "columns": column_headers,
                "style": None,
                "name": gptable.table_name,
            },
        )

    def _smart_write(
        self, row: int, col: int, data: object, format_dict: dict, *args
    ) -> None:
        """
        Depending on the input data, this function will write rich strings or
        use the standard `write()` method. For rich strings, the base format is
        merged with each rich format supplied within data.

        Parameters
        ----------
        row : int
            0-indexed row of cell to write to
        col : int
            0-indexed column of cell to write to
        data : str or list or dict
            Simple string to be written with `format_dict` formatting. Or a
            list of alternating string and dictionary objects. Dictionaries
            specify additional formatting to be applied to the following string
            in the list.
            Dictionaries will be written with (first) key as display text and
            (first) value as URL.
        format_dict : dict
            Dictionary containing base format for the string.

        Returns
        -------
        None
        """
        wb = self._workbook  # Reference to Workbook that contains sheet

        if isinstance(data, list):
            if len(data) == 1:
                data = data[0]
                self._smart_write(row, col, data, format_dict, *args)

            elif any([isinstance(element, FormatList) for element in data]):
                self._write_with_newlines_and_custom_formats(
                    wb, row, col, data, format_dict, *args
                )

            else:
                self._write_with_newlines(wb, row, col, data, format_dict, *args)

        elif isinstance(data, FormatList):
            if len(data.list) == 2:
                text_format = format_dict.copy()
                text_format.update(data.list[0])
                text_data = data.list[1]
                self._smart_write(row, col, text_data, text_format, *args)
            else:
                self._write_with_custom_formats(wb, row, col, data, format_dict, *args)

        elif isinstance(data, dict):
            self._write_dict_as_url(wb, row, col, data, format_dict, *args)

        elif pd.isna(data):
            self.write_blank(row, col, None, wb.add_format(format_dict))

        else:
            # Write handles all other write types dynamically
            self.write(row, col, data, wb.add_format(format_dict), *args)

    def _write_with_newlines_and_custom_formats(
        self, wb: Workbook, row: int, col: int, data: list, format_dict: dict, *args
    ) -> None:
        """
        Take list of FormatList (and str), join with newline characters and smart write
        """
        data_with_newlines = []

        first_element = data.copy()[0]
        if isinstance(first_element, FormatList):
            first_element = first_element.list
        else:
            first_element = [first_element]
        data_with_newlines.extend(first_element)

        for element in data[1:]:
            if isinstance(element, FormatList):
                element = element.list
                element_stings = [item for item in element if isinstance(item, str)]
                first_string = element_stings[0]
                new_string = "\n" + first_string
                element_with_newline = [
                    new_string if item == first_string else item for item in element
                ]
            else:
                element_with_newline = ["\n" + str(element)]
            data_with_newlines.extend(element_with_newline)

        self._write_with_custom_formats(
            wb, row, col, FormatList(data_with_newlines), format_dict, *args
        )

    def _write_with_newlines(
        self, wb: Workbook, row: int, col: int, data: list, format_dict: dict, *args
    ) -> None:
        """
        Take list of str, join with newline character and write
        """
        data_string = "\n".join(data)

        self.write(row, col, data_string, wb.add_format(format_dict), *args)

    def _write_with_custom_formats(
        self,
        wb: Workbook,
        row: int,
        col: int,
        data: "FormatList",
        format_dict: dict,
        *args,
    ) -> None:
        data_with_custom_formats = []
        for item in data.list:
            # Convert dicts to Format (with merge onto base format)
            if isinstance(item, dict):
                rich_format = format_dict.copy()
                rich_format.update(item)
                data_with_custom_formats.append(wb.add_format(rich_format))
            else:
                data_with_custom_formats.append(item)

        data_with_all_formats = []
        for n in range(len(data_with_custom_formats) - 1):
            data_with_all_formats.append(data_with_custom_formats[n])
            if isinstance(data_with_custom_formats[n], str):
                if isinstance(data_with_custom_formats[n + 1], str):
                    data_with_all_formats.append(wb.add_format(format_dict))
        data_with_all_formats.append(data_with_custom_formats[-1])

        self.write_rich_string(
            row, col, *data_with_all_formats, wb.add_format(format_dict), *args
        )

    def _write_dict_as_url(
        self,
        workbook: Workbook,
        row: int,
        col: int,
        data: dict,
        format_dict: dict,
        *args,
    ) -> None:
        url = list(data.values())[0]
        display_text = list(data.keys())[0]

        url_format = format_dict.copy()
        url_format.update(
            {"underline": True, "font_color": "blue"}
        )  # blue == #0000FF - passes WCAG AA contrast check

        self.write_url(
            row, col, url, workbook.add_format(url_format), display_text, *args
        )

    @staticmethod
    def _apply_format(format_table_slice: object, format_dict: dict) -> None:
        """
        Update all cells of a given dataframe slice with the format
        dictionary. Handles dict, series or dataframes.
        """
        if isinstance(format_table_slice, pd.Series):
            (format_table_slice.apply(lambda d: d.update(format_dict)))
        elif isinstance(format_table_slice, pd.DataFrame):
            # Vectorised for 2D
            (format_table_slice.apply(np.vectorize(lambda d: d.update(format_dict))))
        elif isinstance(format_table_slice, dict):
            format_table_slice.update(format_dict)

    def _set_column_widths(self, widths: list) -> None:
        """
        Set the column widths using a list of widths.
        """
        for col_number in range(len(widths)):
            self.set_column_pixels(col_number, col_number, widths[col_number])

    def _calculate_column_widths(
        self, table: pd.DataFrame, formats_table: pd.DataFrame
    ) -> list:
        """
        Calculate Excel column widths using xlsxwriter's cell_autofit_width for each cell,
        and take the maximum per column.

        Parameters
        ----------
        table : pd.DataFrame
            data table to calculate widths from
        formats_table: pd.DataFrame
            formats table to retrieve font size from

        Returns
        -------
        col_widths : list
            width to apply to Excel columns
        """
        cols = table.shape[1]
        col_widths = []
        for col in range(cols):
            cell_widths = []
            for row in range(table.shape[0]):
                cell_val = table.iloc[row, col]
                longest_line = self._get_longest_line(cell_val)
                format_dict = formats_table.iloc[row, col]
                scaling_factor = self._get_scaling_factor(format_dict, longest_line)
                width = ceil(cell_autofit_width(longest_line) * scaling_factor)
                cell_widths.append(width)
            col_widths.append(max(cell_widths) if cell_widths else 0)
        return col_widths

    def _get_scaling_factor(self, format_dict: dict, text: str) -> float:
        """Return scaling factor for width based on font size, bold formatting,
        and capitalisation."""
        font_size = (
            format_dict.get("font_size", 11) if isinstance(format_dict, dict) else 11
        )
        bold = (
            format_dict.get("bold", False) if isinstance(format_dict, dict) else False
        )

        if text and isinstance(text, str):
            num_upper = sum(1 for c in text if c.isupper())
            upper_ratio = num_upper / len(text) if len(text) > 0 else 0
        else:
            upper_ratio = 0
        capitalisation_factor = 1.0 + 0.15 * upper_ratio
        return (font_size / 11) * (1.1 if bold else 1.0) * capitalisation_factor

    def _get_longest_line(self, cell_val: object) -> str:
        """Return the longest line in a cell value split by newline."""
        cell_val_str = self._get_cell_string(cell_val)
        return max(cell_val_str.split("\n"), key=len)

    def _get_cell_string(self, cell_val: object) -> str:
        """Return the contents from any cell value as a string."""
        if isinstance(cell_val, str):
            return cell_val
        elif isinstance(cell_val, (float, int)):
            return str(cell_val)
        elif isinstance(cell_val, dict):
            return "\n".join([self._get_cell_string(k) for k in cell_val.keys()])
        elif isinstance(cell_val, FormatList):
            return self._get_cell_string(cell_val.string)
        elif isinstance(cell_val, list):
            return "\n".join([self._get_cell_string(item) for item in cell_val])
        else:
            return str(cell_val) if cell_val else ""


class GPWorkbook(Workbook):
    """
    Wrapper for an `XlsxWriter.Workbook` object.
    """

    def __init__(self, filename: str = None, options: dict = {}) -> None:
        super(GPWorkbook, self).__init__(filename=filename, options=options)
        self.theme = None
        self._annotations = None
        # Set default theme
        self.set_theme(gptheme)

    def add_worksheet(
        self, name: str = None, gridlines: str = "hide_all"
    ) -> "GPWorksheet":
        """
        Overwrite add_worksheet() to create a GPWorksheet object.

        Parameters
        ----------
        name : str (optional)
            name of the the worksheet to be created
        gridlines : string, optional
            option to hide or show gridlines on worksheets. "show_all" - don't
            hide gridlines, "hide_printed" - hide printed gridlines only, or
            "hide_all" - hide screen and printed gridlines.

        Returns
        -------
        worksheet : gptables.GPWorksheet
            a worksheet object, which supports writing of GPTable objects
        """
        worksheet = super(GPWorkbook, self).add_worksheet(name, GPWorksheet)
        worksheet.theme = self.theme
        worksheet._workbook = self  # Create reference to wb, for formatting

        worksheet.hide_gridlines(
            {"show_all": 0, "hide_printed": 1, "hide_all": 2}[gridlines]
        )

        return worksheet

    def set_theme(self, theme: "Theme") -> None:
        """
        Sets the theme for all GPTable objects written to the Workbook.

        Parameters
        ----------
        theme : gptables.Theme
            a Theme object containing the formatting to be applied to GPTable
            objects written to Worksheets within this Workbook

        Returns
        -------
        None
        """
        if not isinstance(theme, Theme):
            raise TypeError(
                f"`theme` must be a gptables.Theme object, not: {type(theme)}"
            )
        self.theme = theme

    def _update_annotations(self, sheets: dict) -> None:
        ordered_refs = []
        for gptable in sheets.values():
            gptable._set_annotations(self.theme.description_order)
            ordered_refs.extend(gptable._annotations)

        # remove duplicates from ordered_refs and assign to self._annotations
        self._annotations = list(dict.fromkeys(ordered_refs))

    def make_table_of_contents(
        self,
        sheets: dict,
        additional_elements: list = None,
        column_names: list = None,
        table_name: str = None,
        title: str = None,
        subtitles: list = None,
        instructions: str = None,
    ) -> "GPTable":
        """
        Generate table of contents from sheet and optional customisation parameters.

        Parameters
        ----------
        sheets : dict
            mapping worksheet labels to gptables.GPTable objects
        additional_elements : List[str], optional
            additional GPTable elements to display in the contents table. Allowed
            elements are "subtitles", "scope", "source" and "instructions".
        column_names : List[str], optional
            table of contents column names, defaults to
            "Sheet name", "Table description"
        table_name: str, optional
            contents table name, defaults to "contents_table"
        title : str, optional
            table of contents title, defaults to "Table of contents"
        subtitles: List[str], optional
            list of subtitles as strings
        instructions: str, optional
            description of the page layout
            defaults to "This worksheet contains one table."

        Returns
        ------
        gptables.GPTable
        """
        if column_names is None:
            column_names = ["Sheet name", "Table description"]
        if table_name is None:
            table_name = "contents_table"
        if title is None:
            title = "Table of contents"
        if instructions is None:
            instructions = "This worksheet contains one table."

        if additional_elements is not None:
            valid_elements = ["subtitles", "scope", "source", "instructions"]
            if not all(element in valid_elements for element in additional_elements):
                msg = (
                    "Cover `additional_elements` list can only contain"
                    "'subtitles', 'scope', 'source' and 'instructions'"
                )
                raise ValueError(msg)

        contents_dict = {}
        for label, gptable in sheets.items():
            contents_entry = []
            contents_entry.append(self._strip_annotation_references(gptable.title))

            if additional_elements is not None:
                for element in additional_elements:
                    content = getattr(gptable, element)
                    if element == "subtitles":
                        [
                            contents_entry.append(
                                self._strip_annotation_references(element)
                            )
                            for element in content
                        ]
                    else:
                        contents_entry.append(
                            self._strip_annotation_references(content)
                        )

            link = {label: f"internal:'{label}'!A1"}

            contents_dict[label] = [link, contents_entry]

        contents_table = pd.DataFrame.from_dict(
            contents_dict, orient="index"
        ).reset_index(drop=True)

        contents_table.columns = column_names

        return GPTable(
            table=contents_table,
            table_name=table_name,
            title=title,
            subtitles=subtitles,
            instructions=instructions,
            index_columns={2: 0},
        )

    @staticmethod
    def _strip_annotation_references(text: object) -> object:
        """
        Strip annotation references (as $$ $$) from a str or list text element.
        """
        pattern = r"\$\$.*?\$\$"
        if isinstance(text, str):
            no_annotations = re.sub(pattern, "", text)
        elif isinstance(text, FormatList):
            no_annotations = FormatList(
                [
                    re.sub(pattern, "", part) if isinstance(part, str) else part
                    for part in text.list
                ]
            )
        elif isinstance(text, list):  # TODO: this shouldn't get used - check and delete
            no_annotations = [
                re.sub(pattern, "", part) if isinstance(part, str) else part
                for part in text
            ]

        return no_annotations

    def make_notesheet(
        self,
        notes_table: pd.DataFrame,
        table_name: str = None,
        title: str = None,
        instructions: str = None,
    ) -> "GPTable":
        """
        Generate notes table sheets from notes table and optional customisation parameters.

        Parameters
        ----------
        notes_table : pd.DataFrame
            table with notes reference, text and (optional) link columns
        table_name: str, optional
            notes table name, defaults to "notes_table"
        title : str, optional
            notes page title, defaults to "Notes"
        instructions: str, optional
            description of the page layout
            defaults to "This worksheet contains one table."

        Returns
        ------
        gptables.GPTable
        """
        if table_name is None:
            table_name = "notes_table"

        if title is None:
            title = "Notes"

        if instructions is None:
            instructions = "This worksheet contains one table."

        # order notes table by worksheet reference order
        ordered_refs = self._annotations

        order_df = pd.DataFrame({"order": ordered_refs})

        notes = notes_table.copy()
        notes = notes.rename(columns={notes.columns[0]: "order"})

        ordered_notes = order_df.merge(notes, on="order", how="left")

        unreferenced_notes = notes[~notes["order"].isin(ordered_notes["order"])]

        if not unreferenced_notes.empty:
            warnings.warn(
                f"The following notes are not referenced: {list(unreferenced_notes['order'])}"
            )

            ordered_notes = pd.concat([ordered_notes, unreferenced_notes])

        # replace note references with note number
        ordered_notes = (
            ordered_notes.reset_index()
            .rename(columns={"index": "Note number"})
            .drop(columns=["order"])
        )

        # convert from python 0-indexing
        ordered_notes["Note number"] = ordered_notes["Note number"] + 1

        return GPTable(
            table=ordered_notes,
            table_name=table_name,
            title=title,
            instructions=instructions,
            index_columns={},
        )



from pathlib import Path

import pandas as pd

import gptables as gpt

parent_dir = Path(__file__).parents[1]
penguins_data = pd.read_csv(parent_dir / "test/data/penguins.csv")

penguins_table = gpt.GPTable(
    table=penguins_data,
    table_name="penguins_statistics",
    title="The Palmer Penguins Dataset",
    subtitles=["This is the first subtitle", "This is another subtitle"],
    scope="Penguins",
    source="Palmer Station, Antarctica",
)

penguins_sheets = {"Penguins": penguins_table}

penguins_cover = gpt.Cover(
    cover_label="Cover",
    title="Palmer Penguins Dataset",
    intro=[
        "This spreadsheet contains a table of data obtained from the palmerpenguins package",
        "This is intended to be a simple example of how to use the gptables package to create a spreadsheet with a cover sheet and data sheets.",
    ],
    about=[
        "Additional information about your publication can go here",
        [{"bold": True}, "Publication dates"],
        "Date published: 01 January 2025.",
        "Next release: 01 January 2026.",
        [{"bold": True}, "Methodology notes"],
        "Information on methodology can be useful to users of your data",
        [{"bold": True}, "Notes, blank cells and units"],
        "Some cells in the tables refer to notes which can be found in the notes worksheet. Note markers are presented in square brackets, for example: [note 1].",
        "Blank cells indicate no data. An explanation of why there is no data is given in the notes worksheet, see the column headings for which notes you should refer to.",
        "Some column headings give units, when this is the case the units are presented in round brackets to differentiate them from note markers.",
    ],
    contact=[
        "Tel: 01234 567890",
        "Email: [example@email.address](mailto: example@email.address)",
    ],
)

if __name__ == "__main__":
    output_path = parent_dir / "gpt_adding_cover.xlsx"
    gpt.write_workbook(
        filename=output_path,
        sheets=penguins_sheets,
        cover=penguins_cover,
        contentsheet_options={"additional_elements": ["subtitles", "scope"]},
    )
    print("Output written at: ", output_path)



from dataclasses import dataclass
from typing import List

from gptables.core.gptable import FormatList


@dataclass
class Cover:
    """
    Stores cover sheet properties.

    Attributes
    ----------
    title : str
        Cover page title
    intro : List[str, list], optional
        Introductory text
    about : List[str, list], optional
        About/notes text
    contact : List[str, list], optional
        Contact details text
    cover_label : str
        Cover page tab label, defaults to "Cover"
    width: int
        Width of the column, defaults to 85
    """

    def __init__(
        self,
        title: str,
        intro: List = None,
        about: List = None,
        contact: List = None,
        cover_label: str = "Cover",
        width: int = 85,
    ) -> None:

        self.title = title
        self.intro = self._parse_formatting(intro)
        self.about = self._parse_formatting(about)
        self.contact = self._parse_formatting(contact)
        self.cover_label = cover_label
        self.width = width

        # TODO: Add input validation (e.g. empty list)

    @staticmethod
    def _parse_formatting(attribute) -> List:
        """Check attribute for a list. If there is a list then cast the list to a FormatList in attribute.

        Parameters
        ----------
        attribute : List[str, list]

        Returns
        -------
        List[str, FormatList]
        """

        if isinstance(attribute, list):
            attribute = [
                FormatList(text) if isinstance(text, list) else text
                for text in attribute
            ]
        return attribute



from pathlib import Path

import pandas as pd

import gptables as gpt

parent_dir = Path(__file__).parents[1]
penguins_data = pd.read_csv(parent_dir / "test/data/penguins.csv")

penguins_data_1 = penguins_data.iloc[:, :10]
penguins_data_2 = pd.concat(
    [penguins_data.iloc[:, :3], penguins_data.iloc[:, 10:]], axis=1
)


penguins_table_1 = gpt.GPTable(
    table=penguins_data_1,
    table_name="penguins_statistics_1",
    title="The Palmer Penguins Dataset (Sheet 1)",
    subtitles=["This is the first subtitle", "This is another subtitle"],
    scope="Penguins",
    source="Palmer Station, Antarctica",
)

penguins_table_2 = gpt.GPTable(
    table=penguins_data_2,
    table_name="penguins_statistics_2",
    title="The Palmer Penguins Dataset (Sheet 2)",
    subtitles=[
        "This is the first subtitle for sheet 2",
        "Another subtitle for sheet 2",
    ],
    scope="Penguins",
    source="Palmer Station, Antarctica",
)


penguins_sheets = {"Penguins 1": penguins_table_1, "Penguins 2": penguins_table_2}

if __name__ == "__main__":
    output_path = parent_dir / "gpt_adding_datasheets.xlsx"
    gpt.write_workbook(
        filename=output_path,
        sheets=penguins_sheets,
        contentsheet_options={"additional_elements": ["subtitles", "scope"]},
    )
    print("Output written at: ", output_path)



from pathlib import Path

import pandas as pd

import gptables as gpt

# Read data and arrange
parent_dir = Path(__file__).parent

penguin_data = pd.read_csv(parent_dir / "data/penguins.csv")

# Define table elements
table_name = "penguin_statistics"
title = "Penguins"
subtitles = [
    [
        {"font_name": "Chiller"},
        "The",
        {"font_size": 30},
        " first",
        {"font_color": "red"},
        " subtitle",
    ],
    [
        {"bold": True},
        "The",
        {"italic": True},
        " second",
        {"underline": True},
        " subtitle",
    ],
    [{"font_script": 1}, "Ignore", {"font_script": 2}, " this"],
    [{"font_strikeout": True}, "bye", " "],
]  # checking font formatting
units = {key: "mm" for key in range(2, 5)}
scope = "Penguins"
index = {1: 0, 2: 1}

# Additional formatting

additional_formatting = [
    {
        "column": {
            "columns": ["Species", "Island"],
            "format": {"align": "vcenter"},  # checking vertical alignment
            "include_names": False,
        }
    },
    {
        "column": {
            "columns": [0],
            "format": {"indent": 2, "rotation": 90},
            "include_names": True,
        }
    },  # checking alignment formatting
    {
        "row": {
            "rows": -1,
            "format": {
                "bottom": 1,
                "top": 2,
                "bottom_color": "blue",
                "top_color": "yellow",
                "pattern": 6,
                "bg_color": "lime",
                "fg_color": "pink",
            },  # checking pattern formatting
        }
    },
    {
        "row": {
            "rows": 3,
            "format": {
                "num_format": 3,
                "locked": True,
                "left": 3,
                "left_color": "navy",
            },  # checking number & protection formatting
        }
    },
    {
        "row": {
            "rows": 5,
            "format": {
                "num_format": "0.00",
                "right": 2,
                "right_color": "silver",
            },  # checking number formatting
        }
    },
    {
        "cell": {
            "cells": (1, 1),
            "format": {"text_wrap": True},
        }
    },
]

# or just use kwargs
kwargs = {
    "table_name": table_name,
    "title": title,
    "subtitles": subtitles,
    "units": units,
    "scope": scope,
    "source": None,
    "index_columns": index,
    "additional_formatting": additional_formatting,
}

# Define our GPTable
penguin_table = gpt.GPTable(table=penguin_data, **kwargs)

# Use produce workbook to return GPWorkbook
if __name__ == "__main__":
    output_path = parent_dir / "test_additional_formatting_gptable.xlsx"
    wb = gpt.produce_workbook(filename=output_path, sheets={"Penguins": penguin_table})

    # Carry out additional modifications on the GPWorkbook or GPWorksheets
    # This supports all `XlsxWriter` package functionality
    ws = wb.worksheets()[0]
    ws.set_row(0, 30)  # Set the height of the first row

    # Finally use the close method to save the output
    wb.close()
    print("Output written at: ", output_path)
